import asyncio
from typing import Set

from mokr.connection import DevtoolsConnection
from mokr.constants import INPUT_INSERT_TEXT, INPUT_KEY, KEY_DEFINITIONS


class Keyboard():
    def __init__(self, client: DevtoolsConnection) -> None:
        """
        Class to allow sending key events to emulate a keyboard.

        Use `Keyboard.type("text")` to type text normally; this will dispacth
        multiple key events for each character in the given text.

        For finer control, use `Keyboard.down`, `Keyboard.up` and
        `Keyboard.send_character`.

        Special keys and modifier key can be sent. To view all key names
        and their associated event payload, use `Keyboard.key_definitions`.

        Args:
            client (DevtoolsConnection): Remote `DevtoolsConnection` instance.
        """
        self._client = client
        self._modifiers = 0
        self._pressed_keys: Set[str] = set()

    @staticmethod
    def _modifier_bit(key: str) -> int:
        if key == 'Alt':
            return 1
        if key == 'Control':
            return 2
        if key == 'Meta':
            return 4
        if key == 'Shift':
            return 8
        return 0

    @property
    def key_definitions(self) -> dict[str, dict[str, str | int]]:
        """Dictionary of key names and their corresponding event codes."""
        return KEY_DEFINITIONS

    def _key_description_from_string(self, keyString: str) -> dict:
        shift = self._modifiers & 8
        description = {
            'key': '',
            'keyCode': 0,
            'code': '',
            'text': '',
            'location': 0,
        }
        definition = KEY_DEFINITIONS.get(keyString)
        if not definition:
            raise LookupError(f'Unknown key: {keyString}')
        if 'key' in definition:
            description['key'] = definition['key']
        if shift and definition.get('shiftKey'):
            description['key'] = definition['shiftKey']
        if 'keyCode' in definition:
            description['keyCode'] = definition['keyCode']
        if shift and definition.get('shiftKeyCode'):
            description['keyCode'] = definition['shiftKeyCode']
        if 'code' in definition:
            description['code'] = definition['code']
        if 'location' in definition:
            description['location'] = definition['location']
        if len(description['key']) == 1:
            description['text'] = description['key']
        if 'text' in definition:
            description['text'] = definition['text']
        if shift and definition.get('shiftText'):
            description['text'] = definition['shiftText']
        if self._modifiers & ~8:
            description['text'] = ''
        return description

    async def down(
        self,
        key: str,
        text: str | None = None,
    ) -> None:
        """
        Send a "keyDown" event with the given `key`. Does not automatically
        send a "keyUp" event, use `Keyboard.up` for that.

        Modifier keys effect this method, meaning sending
        `Keyboard.down("shift")` and then `Keyboard.down("m")` will type an
        uppercase "M".

        Args:
            key (str): Name of key. See `Keyboard.key_definitions`.
            text (str | None, optional): Force an "input" event to be
                sent, as well. Defaults to None.
        """
        description = self._key_description_from_string(key)
        auto_repeat = description['code'] in self._pressed_keys
        self._pressed_keys.add(description['code'])
        self._modifiers |= self._modifier_bit(description['key'])
        if text is None:
            text = description['text']
        await self._client.send(
            INPUT_KEY,
            {
                'type': 'keyDown' if text else 'rawKeyDown',
                'modifiers': self._modifiers,
                'windowsVirtualKeyCode': description['keyCode'],
                'code': description['code'],
                'key': description['key'],
                'text': text,
                'unmodifiedText': text,
                'autoRepeat': auto_repeat,
                'location': description['location'],
                'isKeypad': description['location'] == 3,
            }
        )

    async def up(self, key: str) -> None:
        """
        Send a "KeyUp" event with the given `key`.

        Args:
            key (str): Name of key. See `Keyboard.key_definitions`.
            text (str | None, optional): Force an "input" event to be
                sent, as well. Defaults to None.
        """
        description = self._key_description_from_string(key)
        self._modifiers &= ~self._modifier_bit(description['key'])
        if description['code'] in self._pressed_keys:
            self._pressed_keys.remove(description['code'])
        await self._client.send(
            INPUT_KEY,
            {
                'type': 'keyUp',
                'modifiers': self._modifiers,
                'key': description['key'],
                'windowsVirtualKeyCode': description['keyCode'],
                'code': description['code'],
                'location': description['location'],
            }
        )

    async def send_character(self, char: str) -> None:
        """
        Send a character `char` into the parent `mokr.frame.Frame`. Does not
        send "keyDown" or "keyUp" event, only sends "keyPress" and "input".

        Modifier keys do not effect this method, meaning sending
        `Keyboard.down("shift")` and then `Keyboard.send_character("m")` will
        type a lowercase "m".

        Args:
            char (str): Character to send, does not accept special characters.
        """
        await self._client.send(INPUT_INSERT_TEXT, {'text': char})

    async def type_text(self, text: str, delay: int | float = 0) -> None:
        """
        Type characters into the page; whatever element is focused will receive
        the sent input.

        Note that modifier keys do not alter text case, meaning sending
        `Keyboard.down("shift")` and typing `Keyboard.type_text("mokr")`
        will not type "MOKR".

        Args:
            text (str): Text to type.
            delay (int | float, optional): Time in milliseconds to wait between
                each character typed. Defaults to 0.
        """
        for char in text:
            if char in KEY_DEFINITIONS:
                await self.press(char, delay=delay)
            else:
                await self.send_character(char)
            if delay:
                await asyncio.sleep(delay / 1000)

    async def press(self, key: str, delay: int | float | None = None) -> None:
        """
        Press the given `key` by sending `Keyboard.down` and `Keyboard.up`
        events, with the given `delay` in between.

        Modifier keys effect this method, meaning sending
        `Keyboard.down("shift")` and then `Keyboard.press("m")` will type an
        uppercase "M".

        Args:
            key (str): Name of key. See `Keyboard.key_definitions`.
            delay (int | float | None, optional): Time in milliseconds to "hold"
                the key down. Defaults to None.
        """
        await self.down(key)
        if delay is not None:
            await asyncio.sleep(delay / 1000)
        await self.up(key)



import json
import logging
import time
from http.client import HTTPException
from urllib.error import URLError
from urllib.request import urlopen

from mokr.exceptions import BrowserError


LOGGER = logging.getLogger(__name__)


def get_ws_endpoint(url) -> str:
    """
    Get the websocket URL for the remote browser at `url`.

    Args:
        url (_type_): Remote browser URL.

    Raises:
        BrowserError: Raised if browser closes while trying to resolve.

    Returns:
        str: Websocket URL from `<url>/json/version` response.
    """
    url = url + '/json/version'
    timeout = time.time() + 30
    while True:
        if time.time() > timeout:
            raise BrowserError('Browser closed unexpectedly:\n')
        try:
            with urlopen(url) as f:
                data = json.loads(f.read().decode())
            break
        except (URLError, HTTPException):
            pass
        time.sleep(0.1)
    return data['webSocketDebuggerUrl']



from mokr.utils.launch import get_ws_endpoint  # noqa
from mokr.utils.remote import (  # noqa
    add_event_listener,
    format_javascript_exception,
    is_javascript_method,
    release_remote_object,
    remove_event_listeners,
    serialize_remote_object,
)



from mokr.constants.info import (  # noqa
    CHROME_VERSION,
    DEFAULT_CHROME_VERSION,
    FIREFOX_BUILD,
    INSTALL_DIR_NAME,
    INSTALL_PATH,
    MOKR_VERSION,
)
from mokr.constants.keys import KEY_DEFINITIONS  # noqa
from mokr.constants.events import (  # noqa
    BROWSER_CLOSE,
    BROWSER_GET_VERSION,
    CLOSE,
    DISCONNECTED,
    DOM_BOX_MODEL,
    DOM_CONTENT_QUADS,
    DOM_DESCRIBE_NODE,
    DOM_INPUT_FILES,
    DOM_LOADED,
    EMULATION_DISABLE_SCRIPT_EXECUTION,
    EMULATION_ENABLE_TOUCH,
    EMULATION_OVERRIDE_BACKGROUND,
    EMULATION_OVERRIDE_METRICS,
    EMULATION_SET_EMULATED_MEDIA,
    ERROR,
    FETCH_AUTH_REQD,
    FETCH_CONTINUE,
    FETCH_CONTINUE_AUTH,
    FETCH_DISABLED,
    FETCH_DOMAIN_REQUEST_SENT,
    FETCH_ENABLED,
    FETCH_FAIL,
    FETCH_FULFILL,
    FETCH_REQUEST_PAUSED,
    FRAME_ATTACHED,
    FRAME_DETACHED,
    FRAME_NAVIGATED,
    FRAME_NAVIGATED_IN_DOC,
    INPUT_INSERT_TEXT,
    INPUT_KEY,
    INPUT_MOUSE,
    INPUT_TOUCH,
    INSPECTOR_TARGET_CRASHED,
    LIFECYCLE_EVENT,
    LOG_ENABLE,
    LOG_ENTRY_ADDED,
    METRICS,
    NETWORK_CACHE_DISABLE,
    NETWORK_DELETE_COOKIES,
    NETWORK_EMULATE_NETWORK_CONDITIONS,
    NETWORK_ENABLE,
    NETWORK_EXTRA_HEADERS,
    NETWORK_GET_ALL_COOKIES,
    NETWORK_GET_COOKIES,
    NETWORK_GET_RESPONSE_BODY,
    NETWORK_LOADING_FAILED,
    NETWORK_LOADING_FINISHED,
    NETWORK_MGR_REQUEST,
    NETWORK_MGR_REQUEST_FAILED,
    NETWORK_MGR_REQUEST_FINISHED,
    NETWORK_MGR_REQUEST_FROM_CACHE,
    NETWORK_MGR_RESPONSE,
    NETWORK_REQUEST,
    NETWORK_REQUEST_FAILED,
    NETWORK_REQUEST_FINISHED,
    NETWORK_REQUEST_SERVED_FROM_CACHE,
    NETWORK_REQUEST_WILL_BE_SENT,
    NETWORK_RESPONSE,
    NETWORK_RESPONSE_RECVD,
    NETWORK_RESPONSE_RECVD_EXTRA,
    NETWORK_SET_COOKIES,
    NETWORK_USER_AGENT_OVERRIDE,
    PAGE_ADD_SCRIPT_TO_EVAL,
    PAGE_BRING_TO_FRONT,
    PAGE_CLOSE,
    PAGE_CONSOLE,
    PAGE_DIALOG,
    PAGE_DOM_LOADED,
    PAGE_ENABLE,
    PAGE_ENABLE_LIFECYCLE_EVENTS,
    PAGE_ERROR,
    PAGE_FRAME_ATTACHED,
    PAGE_FRAME_DETACHED,
    PAGE_FRAME_NAVIGATED,
    PAGE_FRAME_NAVIGATED_IN_DOC,
    PAGE_FRAME_STOPPED_LOADING,
    PAGE_GET_FRAME_TREE,
    PAGE_GET_LAYOUT,
    PAGE_GET_NAVIGATION_HISTORY,
    PAGE_HANDLE_DIALOG,
    PAGE_JAVASCRIPT_DIALOG_OPEN,
    PAGE_LIFECYCLE_EVENT,
    PAGE_LOAD,
    PAGE_LOAD_EVENT_FIRED,
    PAGE_NAVIGATE,
    PAGE_NAVIGATE_TO_HISTORY_ENTRY,
    PAGE_PRINT_TO_PDF,
    PAGE_RELOAD,
    PAGE_SCREENSHOT,
    PAGE_SET_BYPASS_CSP,
    PERFORMANCE_ENABLE,
    PERFORMANCE_GET_METRICS,
    PERFORMANCE_METRICS,
    RUMTIME_BINDING_CALL,
    RUNTIME_ADD_BINDING,
    RUNTIME_CALL_FUNCTION,
    RUNTIME_CONSOLE_API_CALL,
    RUNTIME_ENABLE,
    RUNTIME_EVALUATE,
    RUNTIME_EXCEPTION_THROWN,
    RUNTIME_EXECUTION_CONTEXT_CREATED,
    RUNTIME_EXECUTION_CONTEXT_DESTROYED,
    RUNTIME_EXECUTION_CONTEXTS_CLEARED,
    RUNTIME_GET_PROPERTIES,
    RUNTIME_QUERY_OBJECTS,
    RUNTIME_RELEASE_OBJECT,
    SECURITY_ENABLE,
    SECURITY_IGNORE_HTTPS_ERRORS,
    TARGET_ACTIVATE,
    TARGET_ATTACH,
    TARGET_ATTACHED,
    TARGET_CHANGED,
    TARGET_CLOSE,
    TARGET_CREATE_BROWSER_CONTEXT,
    TARGET_CREATE_TARGET,
    TARGET_CREATED,
    TARGET_DESTROYED,
    TARGET_DETACHED,
    TARGET_DISPOSE_BROWSER_CONTEXT,
    TARGET_GET_CONTEXTS,
    TARGET_INFO_CHANGED,
    TARGET_RECV_MSG,
    TARGET_SEND_DETACH,
    TARGET_SEND_MSG,
    TARGET_SET_AUTO_ATTACH,
    TARGET_SET_DISCOVER_TARGETS,
    TARGET_TARGET_CREATED,
    TARGET_TARGET_DESTROYED,
    WORKER_CREATED,
    WORKER_DESTROYED,
)
from mokr.constants.javascript import (  # noqa
    METHOD_ADD_PAGE_BINDING,
    METHOD_DELIVER_BINDING_RESULT,
    METHOD_ELEMENT_IN_VIEW,
    METHOD_EMBED_JAVASCRIPT_BY_CONTENT,
    METHOD_EMBED_JAVASCRIPT_BY_URL,
    METHOD_EMBED_STYLE_BY_CONTENT,
    METHOD_EMBED_STYLE_BY_URL,
    METHOD_EVAL_XPATH,
    METHOD_FETCH_REQUEST,
    METHOD_FOCUS_ELEMENT,
    METHOD_GET_CONTENT,
    METHOD_GET_PROPERTY,
    METHOD_SCROLL_INTO_VIEW,
    METHOD_SELECT_VALUES,
    METHOD_SET_CONTENT,
    METHOD_WAIT_FOR_PREDICATE_PAGE,
    METHOD_WAIT_FOR_XPATH_OR_SELECTOR,
)
from mokr.constants.typing import (  # noqa
    HTTP_CACHE_TYPES,
    HTTP_METHODS,
    LIFECYCLE_EVENTS,
    REFERRER_POLICIES,
)
from mokr.constants.http import (  # noqa
    HTTP_RESPONSE_CODES_TEXT_MAP,
    NETWORK_ERROR_CODES_TO_REASONS,
)



from __future__ import annotations

import logging
import math
import re
from typing import TYPE_CHECKING, Any

from mokr.connection import DevtoolsConnection
from mokr.constants import (
    RUNTIME_CALL_FUNCTION,
    RUNTIME_EVALUATE,
    RUNTIME_QUERY_OBJECTS,
)
from mokr.exceptions import ElementHandleError, NetworkError
from mokr.execution.handle.javascript import JavascriptHandle
from mokr.utils.remote import format_javascript_exception, is_javascript_method

if TYPE_CHECKING:
    from mokr.frame import Frame


LOGGER = logging.getLogger(__name__)

EVALUATION_SCRIPT_URL = '__mokr_evaluation_script__'

SOURCE_URL_REGEX = re.compile(
    r'^[\040\t]*//[@#] sourceURL=\s*(\S*?)\s*$',
    re.MULTILINE,
)


class ExecutionContext():
    def __init__(
        self,
        client: DevtoolsConnection,
        context_payload: dict,
        object_handle_factory: Any,
        frame: Frame = None,
    ) -> None:
        """
        Class to handle execution of JavaScript functions, element manipulation,
        and associated runtime error translation.

        Args:
            client (DevtoolsConnection): Remote connection from parent.
            context_payload (dict): Context from triggering event.
            object_handle_factory (Any): Factory for handling remote objects.
                Should create `mokr.execution.JavascriptHandle` based objects.
            frame (Frame, optional): The `mokr.frame.Frame` that spawned this.
                Defaults to None.
        """
        self._client = client
        self._frame = frame
        self._context_id = context_payload.get('id')
        aux_data = context_payload.get('auxData', {'isDefault': False})
        self._is_default = bool(aux_data.get('isDefault'))
        self._object_handle_factory = object_handle_factory

    @property
    def frame(self) -> Frame | None:
        """Parent `mokr.frame.Frame` that spawned this."""
        return self._frame

    @staticmethod
    def _rewrite_exception(exception: Exception) -> None:
        if exception.args[0].endswith('Cannot find context with specified id'):
            msg = 'Execution context destroyed, likely due to a navigation.'
            raise type(exception)(msg)
        raise exception

    def _convert_argument(self, arg: Any) -> dict:
        if arg == math.inf:
            return {'unserializableValue': 'Infinity'}
        if arg == -math.inf:
            return {'unserializableValue': '-Infinity'}
        object_handle = arg if isinstance(arg, JavascriptHandle) else None
        if object_handle:
            if object_handle._context != self:
                raise ElementHandleError(
                    'JavascriptHandle can only be evaluated'
                    ' in the context it was created in.'
                )
            if object_handle._disposed:
                raise ElementHandleError('JavascriptHandle is disposed!')
            if object_handle._remote_object.get('unserializableValue'):
                return {
                    'unserializableValue': object_handle._remote_object.get(
                        'unserializableValue'
                    )
                }
            if not object_handle._remote_object.get('objectId'):
                return {'value': object_handle._remote_object.get('value')}
            return {'objectId': object_handle._remote_object.get('objectId')}
        return {'value': arg}

    async def evaluate(
        self,
        page_function: str,
        *args: Any,
        force_expr: bool = False,
    ) -> Any:
        """
        Run `ExecutionContext.evaluate_handle` and try to run
        `mokr.execution.JavascriptHandle.json` on the result to return the
        object as a dictionary.

        Args:
            page_function (str): JavaScript function to run.
            force_expr (bool): If True, treat `page_function` as an expression.
                Otherwise, automatically determine if it is a function or
                an expression. Defaults to False.

        Raises:
            `mokr.exceptions.NetworkError` if an unhandled error occurs
            either evaluating the function or requesting the resulting object.

        Returns:
            dict | None: The decoded object (dict) via
                `mokr.execution.JavascriptHandle.json` or None if a known
                error occurs decoding it.
        """
        handle = await self.evaluate_handle(
            page_function,
            *args,
            force_expr=force_expr,
        )
        try:
            result = await handle.json()
        except NetworkError as e:
            if any(
                error_part in e.args[0] for error_part in [
                    'Object reference chain is too long',
                    'Object couldn\'t be returned by value',
                ]
            ):
                return None
            raise
        await handle.dispose()
        return result

    async def evaluate_handle(
        self,
        page_function: str,
        *args: Any,
        force_expr: bool = False,
        eval_script_url_suffix: str | None = None,
    ) -> JavascriptHandle:
        """
        Execute a JavaScript function with given arguments.

        Args:
            page_function (str): JavaScript function to run.

        Raises:
            ElementHandleError: Raised if execution fails.

        Returns:
            JavascriptHandle: `mokr.execution.JavascriptHandle`.
        """
        suffix = f'//# sourceURL={EVALUATION_SCRIPT_URL}'
        if eval_script_url_suffix:
            suffix += eval_script_url_suffix
        try:
            if force_expr or (
                not args and not is_javascript_method(page_function)
            ):
                if SOURCE_URL_REGEX.match(page_function):
                    expression_with_source_url = page_function
                else:
                    expression_with_source_url = f'{page_function}\n{suffix}'
                _obj = await self._client.send(
                    RUNTIME_EVALUATE,
                    {
                        'expression': expression_with_source_url,
                        'context_id': self._context_id,
                        'returnByValue': False,
                        'awaitPromise': True,
                        'userGesture': True,
                    },
                )
            else:
                _obj = await self._client.send(
                    RUNTIME_CALL_FUNCTION,
                    {
                        'functionDeclaration': f'{page_function}\n{suffix}\n',
                        'executionContextId': self._context_id,
                        'arguments': [
                            self._convert_argument(arg) for arg in args
                        ],
                        'returnByValue': False,
                        'awaitPromise': True,
                        'userGesture': True,
                    },
                )
        except Exception as e:
            self._rewrite_exception(e)
        exception_details = _obj.get('exceptionDetails')
        if exception_details:
            js_exception = format_javascript_exception(exception_details)
            raise ElementHandleError(f'Evaluation failed: {js_exception}')
        remote_object = _obj.get('result')
        return self._object_handle_factory(remote_object)

    async def query_objects(
        self,
        javascript_handle: JavascriptHandle,
    ) -> JavascriptHandle:
        """
        Query all objects with the given `mokr.execution.JavascriptHandle`'s
        `get_property("objectId")`.

        Args:
            javascript_handle (JavascriptHandle): A valid (not disposed)
                `mokr.execution.JavascriptHandle` object.

        Raises:
            ElementHandleError: Raised if the given handle is disposed, or
                does not have an "objectId" remote property (primitive type).

        Returns:
            JavascriptHandle: A `mokr.execution.JavascriptHandle` initialised
                from the remot response.
        """
        if javascript_handle._disposed:
            raise ElementHandleError('Prototype JavascriptHandle is disposed.')
        if not javascript_handle._remote_object.get('objectId'):
            raise ElementHandleError(
                'Prototype JavascriptHandle cannot reference primitive value.'
            )
        response = await self._client.send(
            RUNTIME_QUERY_OBJECTS,
            {'prototypeObjectId': javascript_handle._remote_object['objectId']}
        )
        return self._object_handle_factory(response.get('objects'))



import asyncio
import logging
from typing import Literal

from mokr.browser import Browser
from mokr.connection import Connection
from mokr.constants import BROWSER_CLOSE, MOKR_VERSION, TARGET_GET_CONTEXTS
from mokr.launch import ChromeLauncher, FirefoxLauncher
from mokr.utils import get_ws_endpoint


version = MOKR_VERSION
version_info = tuple(int(i) for i in version.split('.'))


def launch(
    browser_type: Literal["chrome", "firefox"] = "chrome",
    binary_path: str = None,
    headless: bool = None,
    user_data_dir: str = None,
    devtools: bool = False,
    ignore_default_args: bool | list[str] = False,
    ignore_https_errors: bool = False,
    default_viewport: dict[str, int] = None,
    proxy: str = None,
    default_user_agent: str = None,
    slow_mo: int = 0,
    log_level: str | int = None,
    args: list[str] = None,
    dumpio: bool = False,
    env: dict[str, str] = None,
    loop: asyncio.AbstractEventLoop = None,
    firefox_user_prefs: dict = None,
    firefox_addons_paths: list[str] = None,
) -> Browser:
    """
    Launch a browser process and create a `mokr.browser.Browser`.
    Wrapper for `mokr.launch.Launcher.launch`.

    Args:
        browser_type (Literal["chrome", "firefox"]): The type of browser to,
            use. One of "chrome" or "firefox". Note that Firefox is not fully
            implemented and only offers partial functionality.
        binary_path (str, optional): Path to executable to use.
            Defaults to None (looks for default executable that can be installed
            via `mokr install`).
        headless (bool, optional): Run the browser in headless (no window) mode.
            Defaults to None (uses opposite value to `devtools`).
        user_data_dir (str, optional): Path to a user data directory.
            Defaults to None.
        devtools (bool, optional): Automatically open the developer tools panel.
            Defaults to False.
        ignore_default_args (bool | list[str], optional): Either a bool to
            indicate ignoring all arguments or a list of arguments to
            ignore.
            Be cautious, ignoring some arguments may cause unexpected results.
            Defaults to False.
        ignore_https_errors (bool, optional): Ignore site security errors.
            Defaults to False.
        default_viewport (dict[str, int], optional): Set the default viewport
            for new pages. Accepts a dictionary keyed with viewport options.
            Not all viewport options are considered, only: "isMobile", "width",
            "height", "deviceScaleFactor", "isLandscape", and "hasTouch".
            Defaults to None (800x600 viewport).
        proxy (str, optional): Proxy to route all requests through. Can
            be a regular HTTP/S proxy or SOCKS proxy. Expects proxy as
            `<scheme>://[(optional)<username>:<password>]@<host><password>`.
        default_user_agent (str, optional): Default user agent to use on
            all new pages.
        slow_mo (int, optional): Slow execution of remote calls by the given
            time in milliseconds. Defaults to 0.
        log_level (str | int, optional): Log level to log at.
            Defaults to None (same as root).
        args (list[str], optional): Additional arguments to pass to the
            browser process when launching. Defaults to None.
        dumpio (bool, optional): Pipe the browser process' stdout and stderr
            into ``process.stdout`` and ``process.stderr``. Defaults to False.
        env (dict[str, str], optional): Additional environment variables that
            the browser process will be able to read. Defaults to None.
        loop (asyncio.AbstractEventLoop, optional): A running asyncio loop
            to execute within. Defaults to None (uses `asyncio.get_event_loop`).
        firefox_user_prefs (dict): Firefox only. User preferences to load.
        firefox_addons_paths (list[str]): Firefox only. A list of paths to
            addons that will be installed as temporary extensions.

    Example::

        ```python
        from mokr import launch

        async with launch() as browser:
            page = await browser.first_page()
            await page.goto("https://example.com")

        # Or, to avoid the contextmanager.
        launcher = launch()
        browser = await launcher.launch()
        page = await browser.first_page()
        await page.goto("https://example.com")
        await launcher.stop()
        ```

    Raises:
        ValueError: Raised if `browser_type` isn't of "chrome" or "firefox".

    Returns:
        Browser: A newly created `mokr.browser.Browser` instance.
    """
    launcher_classes = {
        "chrome": ChromeLauncher,
        "firefox": FirefoxLauncher,
    }
    launcher_class = launcher_classes.get(browser_type.lower())
    if not launcher_class:
        raise ValueError(f"Invalid browser type given: {browser_type}")
    return launcher_class(
        binary_path,
        headless,
        user_data_dir,
        devtools,
        ignore_default_args,
        ignore_https_errors,
        default_viewport,
        proxy,
        default_user_agent,
        slow_mo,
        log_level,
        args,
        dumpio,
        env,
        loop,
        firefox_user_prefs,
        firefox_addons_paths,
    )


async def connect(
    browser_type: Literal["chrome", "firefox"] = "chrome",
    browser_ws_endpoint: str = None,
    browser_url: str = None,
    ignore_https_errors: bool = False,
    default_viewport: dict[str, int] = None,
    slow_mo: int = 0,
    log_level: str | int = None,
    loop: asyncio.AbstractEventLoop = None,
) -> Browser:
    """
    Connect to an existing running browser.

    Args:
        browser_type (Literal["chrome", "firefox"]): The type of browser to
            connect to. One of "chrome" or "firefox". Note that Firefox is not
            fully implemented and only offers partial functionality.
        browser_ws_endpoint (str, optional): An existing browser websocket
            endpoint to connect to. Should be formated like
            `"ws://${host}:${port}/devtools/browser/<id>"`.
            Defaults to None, if not given, must give `browser_url`.
        browser_url (str, optional): An existing browser URL to connect to and
            get the websocket URL from.  Should follow format of
            "http://${host}:${port}".
            Defaults to None, if not given, must give `browser_ws_endpoint`.
        ignore_https_errors (bool, optional): Ignore site security errors.
            Defaults to False.
        default_viewport (dict[str, int], optional): Set the default viewport
            for new pages. Accepts a dictionary keyed with viewport options.
            Not all viewport options are considered, only: "isMobile", "width",
            "height", "deviceScaleFactor", "isLandscape", and "hasTouch".
            Defaults to None (800x600 viewport).
        slow_mo (int, optional): Slow execution of remote calls by the given
            time in milliseconds. Defaults to 0.
        log_level (str | int, optional): Log level to log at.
            Defaults to None (same as root).
        loop (asyncio.AbstractEventLoop, optional): A running asyncio loop
            to execute within. Defaults to None (uses `asyncio.get_event_loop`).

    Raises:
        ValueError: Raised if `browser_type` isn't of "chrome" or "firefox" or
            neither `browser_ws_endpoint` nor `browser_url` are given.

    Returns:
        Browser: A newly created `mokr.browser.Browser` instance.
    """
    if log_level is not None:
        logging.getLogger('mokr').setLevel(log_level)
    if browser_type not in ("chrome", "firefox"):
        raise ValueError(f"Invalid browser type given: {browser_type}")
    if not browser_ws_endpoint:
        if not browser_url:
            raise ValueError(
                'Must give one of browser_ws_endpoint or browser_url.'
            )
        browser_ws_endpoint = get_ws_endpoint(browser_url)
    connection = Connection(
        browser_ws_endpoint,
        loop if loop else asyncio.get_event_loop(),
        slow_mo,
    )
    browser_context_ids = (
        await connection.send(TARGET_GET_CONTEXTS)
    ).get('browserContextIds', [])
    default_viewport = (
        default_viewport if default_viewport
        else {'width': 800, 'height': 600}
    )
    browser = Browser(
        browser_type,
        connection,
        browser_context_ids,
        ignore_https_errors,
        default_viewport,
        None,
        lambda: connection.send(BROWSER_CLOSE),
    )
    await browser.start()
    return browser



from mokr.connection import DevtoolsConnection
from mokr.constants import PAGE_HANDLE_DIALOG


class Dialog():
    def __init__(
        self,
        client: DevtoolsConnection,
        dialog_type: str,
        message: str,
        default_value: str | None = None,
    ) -> None:
        """
        Dialog objects are initialised by the parent `mokr.browser.Page`
        when a dialog event is triggered.

        Args:
            client (DevtoolsConnection): A `mokr.connection.DevtoolsConnection`
                spawned by the parent `mokr.browser.Page`.
            dialog_type (str): Dialog type from the triggering event's type.
            message (str): Message in the dialog.
            default_value (str | None, optional): Default prompt that
                the dialog spawned with. Defaults to None.

        Example::

            async def close_dialog(dialog):
                print(dialog.message)
                await dialog.dismiss()
            page.on("dialog", close_dialog)
        """
        self._client = client
        self._type = dialog_type
        self._message = message
        self._handled = False
        self._default_value = '' if default_value is None else default_value

    @property
    def kind(self) -> str:
        """
        Get the remote `Dialog` event type. One of "alert", "beforeunload,
        "confirm", "prompt, or "".
        """
        return self._type

    @property
    def message(self) -> str:
        """Get the message the dialog spawned with."""
        return self._message

    @property
    def default_value(self) -> str:
        """
        Get default prompt value, if remote dialog type is "prompt".
        Otherwise, get "".
        """
        return self._default_value

    async def accept(self, prompt_text: str = '') -> None:
        """
        Accept the remote dialog. Can optionally accept with the given
        `prompt_text`, if `Dialog.type` is "prompt".

        Args:
            prompt_text (str, optional): _description_. Defaults to ''.
        """
        self._handled = True
        await self._client.send(
            PAGE_HANDLE_DIALOG,
            {
                'accept': True,
                'prompt_text': prompt_text,
            }
        )

    async def dismiss(self) -> None:
        """Dismiss the remote dialog."""
        self._handled = True
        await self._client.send(
            PAGE_HANDLE_DIALOG,
            {'accept': False}
        )



from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Any, Generator, Literal

from mokr.constants import METHOD_WAIT_FOR_PREDICATE_PAGE
from mokr.exceptions import MokrTimeoutError, NetworkError, PageError
from mokr.utils.remote import is_javascript_method

if TYPE_CHECKING:
    from mokr.frame.frame import Frame


LOGGER = logging.getLogger(__name__)


class WaitTask():
    def __init__(
        self,
        frame: Frame,
        predicate_body: str,
        title: str,
        polling: Literal['raf', 'mutation'] | int | float,
        timeout: float,
        loop: asyncio.AbstractEventLoop,
        *args: Any,
    ) -> None:
        """
        Class used to monitor a remote frame for events to occur within
        a given timer interval.

        Args:
            frame (Frame): `mokr.frame.Frame` that spawned this `WaitTask`.
            predicate_body (str): JavaScript function to return from.
            title (str): Title to be used to contextualise error, if any.
            polling (Literal["raf", "mutation"] | int | float): Polling type;
                if set to "raf", executes continously in
                "requestAnimationFrame", else if set to "mutation" executes
                only on DOM mutations.
                Defaults to "raf".
            timeout (float): Time in milliseconds to wait.
            loop (asyncio.AbstractEventLoop): Running asyncio loop.

        Raises:
            ValueError: Raised if `polling_type` not "raf" or "mutation".
        """
        polling_value_error = False
        if isinstance(polling, str):
            if polling not in ['raf', 'mutation']:
                polling_value_error = True
        elif isinstance(polling, (int, float)):
            if polling <= 0:
                polling_value_error = True
        else:
            polling_value_error = True
        if polling_value_error:
            raise ValueError(
                "Unsupported polling value, must provide one of"
                '"raf", "mutation", or a positive int.'
                f" Provided: {polling}"
            )
        self._frame = frame
        self._polling = polling
        self._timeout = timeout
        self._loop = loop
        self._args = args
        self._run_count = 0
        self._terminated = False
        self._timeout_error = False
        if args or is_javascript_method(predicate_body):
            self._predicate_body = f'return ({predicate_body})(...args)'
        else:
            self._predicate_body = f'return {predicate_body}'
        frame._wait_tasks.add(self)
        self.promise = self._loop.create_future()
        if timeout:
            self._timeout_timer = self._loop.create_task(
                self._timer(title, self._timeout),
            )
        self._running_task = self._loop.create_task(self.rerun())

    def __await__(self) -> Generator:
        result = yield from self.promise
        if isinstance(result, Exception):
            raise result
        return result

    def _cleanup(self) -> None:
        if self._timeout and not self._timeout_error:
            self._timeout_timer.cancel()
        self._frame._wait_tasks.remove(self)

    async def _timer(self, title: str, timeout: int | float) -> None:
        await asyncio.sleep(timeout / 1000)
        self._timeout_error = True
        self.terminate(
            MokrTimeoutError(
                f'Waiting for {title} failed: timeout {timeout}ms exceeds.'
            )
        )

    def terminate(self, error: Exception) -> None:
        """
        Finish the task, if the promise is still not done, set the result
        with the given `error`, then remove this task from the parent
        `mokr.frame.Frame`.

        Args:
            error (Exception): Error to raise if watched promise isn't done.
        """
        self._terminated = True
        if not self.promise.done():
            self.promise.set_result(error)
        self._cleanup()

    async def rerun(self) -> None:
        """
        Start polling for the expected condition.

        Raises:
            PageError: Raised if no `mokr.execution.ExecutionContext` attached
                to the parent `mokr.frame.Frame`.
        """
        run_count = self._run_count = self._run_count + 1
        success = None
        error = None
        try:
            context = await self._frame.execution_context()
            if context is None:
                raise PageError('No execution context attached to frame.')
            success = await context.evaluate_handle(
                METHOD_WAIT_FOR_PREDICATE_PAGE,
                self._predicate_body,
                self._polling,
                self._timeout,
                *self._args,
            )
        except Exception as e:
            error = e
        if self.promise.done():
            return
        if self._terminated or run_count != self._run_count:
            if success:
                await success.dispose()
            return
        # Add try/except referring to puppeteer.
        try:
            if not error and success and (
                await self._frame.evaluate('s => !s', success)
            ):
                await success.dispose()
                return
        except NetworkError:
            if success is not None:
                await success.dispose()
            return
        # Page is navigated and context is destroyed.
        # Try again in the new execution context.
        if isinstance(error, NetworkError) and any(
            error_part in error.args[0] for error_part in [
                'Execution context was destroyed',
                'Cannot find context with specified id',
            ]
        ):
            return
        if error:
            self.promise.set_exception(error)
        else:
            self.promise.set_result(success)
        self._cleanup()



import asyncio
from typing import Literal

from mokr.connection import DevtoolsConnection
from mokr.constants import INPUT_MOUSE
from mokr.input.keyboard import Keyboard


class Mouse():
    def __init__(self, client: DevtoolsConnection, keyboard: Keyboard) -> None:
        """
        Class to emulate mouse movement.

        The `Mouse`'s remote position is measured in pixels, with origin at
        the top-left corner of the viewport.

        Args:
            client (DevtoolsConnection): Remote `DevtoolsConnection` instance.
            keyboard (Keyboard): Active `mokr.input.Keyboard` from the parent
                `mokr.input.Keyboard`. Passes active modifiers such as CTRL.
        """
        self._client = client
        self._keyboard = keyboard
        self._x = 0.0
        self._y = 0.0
        self._button = 'none'

    async def move(
        self,
        x: float,
        y: float,
        steps: int = 1,
    ) -> None:
        """
        Move the mouse to target coordinates (`x`, `y`), sending intermittent
        events along the trail for each step in `steps`.

        Args:
            x (float): Target X coordinate.
            y (float): Target Y coordinate.
            steps (int, optional): Number of times to stop along the path.
                Defaults to 1 (final destination only).
        """
        start_x = self._x
        start_y = self._y
        self._x = x
        self._y = y
        for i in range(1, steps + 1):
            x = round(start_x + (self._x - start_x) * (i / steps))
            y = round(start_y + (self._y - start_y) * (i / steps))
            await self._client.send(
                INPUT_MOUSE,
                {
                    'type': 'mouseMoved',
                    'button': self._button,
                    'x': x,
                    'y': y,
                    'modifiers': self._keyboard._modifiers,
                },
            )

    async def down(
        self,
        button: Literal["left", "right", "middle"] = "left",
        click_count: int = 1,
    ) -> None:
        """
        Send a "mousePressed" event with given mouse `button`, repeat for
        `click_count` number of times.

        Args:
            button (Literal["left", "right", "middle"], optional): Mouse button
                to send with. Defaults to "left".
            click_count (int, optional): Number of click events to
                send. Defaults to 1.
        """
        self._button = button
        await self._client.send(
            INPUT_MOUSE,
            {
                'type': 'mousePressed',
                'button': self._button,
                'x': self._x,
                'y': self._y,
                'modifiers': self._keyboard._modifiers,
                'clickCount': click_count,
            }
        )

    async def up(
        self,
        button: Literal["left", "right", "middle"] = "left",
        click_count: int = 1,
    ) -> None:
        """
        Send a "mouseReleased" event with given mouse `button`, repeat for
        `click_count` number of times.

        Args:
            button (Literal["left", "right", "middle"], optional): Mouse button
                to send with. Defaults to "left".
            click_count (int, optional): Number of click events to
                send. Defaults to 1.
        """
        self._button = button
        await self._client.send(
            INPUT_MOUSE,
            {
                'type': 'mouseReleased',
                'button': self._button,
                'x': self._x,
                'y': self._y,
                'modifiers': self._keyboard._modifiers,
                'clickCount': click_count,
            },
        )

    async def click(
        self,
        x: float,
        y: float,
        button: Literal["left", "right", "middle"] = "left",
        click_count: int = 1,
        delay: int | float | None = 1000,
    ) -> None:
        """
        Move the pointer to the target coordinates and click.

        Shortcut to running `Mouse.move` then `Mouse.down` then `Mouse.up`.

        Args:
            x (float): X coordinate to move to.
            y (float): Y coordinate to move to.
            button (Literal["left", "right", "middle"], optional): Mouse button
                to click with. Defaults to "left".
            click_count (int, optional): Number of clicks to run. Defaults to 1.
            delay (int | float | None, optional): Time in milliseconds to wait
                before each click. Defaults to 1000.
        """
        await self.move(x, y)
        await self.down(button, click_count)
        if delay is not None:
            await asyncio.sleep(delay / 1000)
        await self.up(button, click_count)



from __future__ import annotations

import asyncio
import logging
from asyncio import Future
from typing import TYPE_CHECKING, Awaitable

from pyee import EventEmitter

from mokr.connection.base import RemoteConnection
from mokr.constants import TARGET_SEND_DETACH, TARGET_SEND_MSG
from mokr.exceptions import NetworkError

if TYPE_CHECKING:
    from mokr.connection.connection import Connection


LOGGER = logging.getLogger(__name__)


class DevtoolsConnection(EventEmitter, RemoteConnection):
    def __init__(
        self,
        connection: Connection | DevtoolsConnection,
        target_type: str,
        session_id: str,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        """
        Create new CDP (Chrome Devtools Protocol) session.

        Most functionality is controlled by calling CDP methods or handling
        CDP events. Read more about these at:
        https://chromedevtools.github.io/devtools-protocol/

        Args:
            connection (Connection | DevtoolsConnection): Parent remote
                connection to use.
            target_type (str): Type of target this connection is for.
            session_id (str): Unique session identifier.
            loop (asyncio.AbstractEventLoop): Running asyncio loop.
        """
        super().__init__()
        self._last_id = 0
        self._callbacks: dict[int, asyncio.Future] = {}
        self._connection: Connection | None = connection
        self._target_type = target_type
        self._sessionId = session_id
        self._sessions: dict[str, DevtoolsConnection] = dict()
        self._loop = loop

    def _on_successful_response(self, callback: Future, msg: dict) -> None:
        if callback and not callback.done():
            callback.set_result(msg.get('result'))

    def _on_query(self, msg: dict) -> None:
        _, method, params = self._handle_detached_or_received(msg)
        self.emit(method, params)

    def _on_closed(self) -> None:
        for cb in self._callbacks.values():
            cb.set_exception(self._rewrite_exception(
                cb.error,
                f'Protocol error {cb.method}: Target closed.',
            ))
        self._callbacks.clear()
        self._connection = None

    def _create_session(
        self,
        target_type: str,
        session_id: str,
    ) -> DevtoolsConnection:
        session = DevtoolsConnection(self, target_type, session_id, self._loop)
        self._sessions[session_id] = session
        return session

    def send(self, method: str, params: dict = None) -> Awaitable[dict]:
        """
        Send a message to the remote connection.

        Args:
            method (str): Method to run.
            params (dict, optional): Arguments for method, if any.
                Defaults to None.

        Raises:
            NetworkError: Raised if connection is closed.

        Returns:
            Awaitable[dict]: Remote response as dictionary.
        """
        if not self._connection:
            raise NetworkError(
                f'Protocol Error ({method}): Session closed. Most likely the'
                f' {self._target_type} has been closed.'
            )
        msg = self._prepare_message(method, params)
        callback = self._loop.create_future()
        self._callbacks[self._last_id] = callback
        callback.error = NetworkError()
        callback.method = method
        try:
            self._connection.send(
                TARGET_SEND_MSG,
                {
                    'sessionId': self._sessionId,
                    'message': msg,
                }
            )
        except Exception as e:
            # The response from target may have already been dispatched.
            if self._last_id in self._callbacks:
                _callback = self._callbacks.pop(self._last_id)
                _callback.set_exception(
                    self._rewrite_exception(
                        _callback.error,
                        e.args[0],
                    )
                )
        return callback

    async def detach(self) -> None:
        """
        Detach session from it's target. Once detached, it is invalid and
        cannot receive or send messages.

        Raises:
            NetworkError: Raised if remote connection is already closed.
        """
        if not self._connection:
            raise NetworkError('Connection already closed.')
        await self._connection.send(
            TARGET_SEND_DETACH,
            {'sessionId': self._sessionId},
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Callable

from pyee import EventEmitter

from mokr.connection import DevtoolsConnection
from mokr.constants import (
    NETWORK_CACHE_DISABLE,
    NETWORK_EMULATE_NETWORK_CONDITIONS,
    NETWORK_ENABLE,
    NETWORK_EXTRA_HEADERS,
    NETWORK_USER_AGENT_OVERRIDE,
    SECURITY_IGNORE_HTTPS_ERRORS,

)
from mokr.frame import FrameManager
from mokr.execution.context import EVALUATION_SCRIPT_URL

if TYPE_CHECKING:
    from mokr.browser.page import Page


class NetworkManager(EventEmitter):
    def __init__(
        self,
        page: Page,
        client: DevtoolsConnection,
        frame_manager: FrameManager,
        interception_callback_chain: list[Callable],
    ) -> None:
        super().__init__()
        self._page = page
        self._client = client
        self._frame_manager = frame_manager
        self._interception_callback_chain = interception_callback_chain
        self._extra_http_headers = {}
        self._protocol_request_interception_enabled = None
        self._user_request_interception_enabled = True
        self._user_cache_disabled = True
        self._attempted_authentications = set()
        self._emulated_network_conditions = {}
        self.credentials = {}
        self.user_agent = ''
        self.user_agent_metadata = ''
        for event, method in self._events_to_methods.items():
            self._client.on(event, method)

    @classmethod
    async def create(
        cls,
        page: Page,
        client: DevtoolsConnection,
        frame_manager: FrameManager,
        ignore_https_errors: bool,
        interception_callback_chain: list[Callable]
    ) -> NetworkManager:
        """
        Async constructor for this class. Necessary to run some asyncronous
        post-initialisation tasks.

        Args:
            page (Page): Parent `mokr.browser.Page`.
            client (DevtoolsConnection): A `mokr.connection.DevtoolsConnection`
                spawned by the parent `mokr.browser.Page`.
            frame_manager (FrameManager): A `mokr.frame.FrameManager`
                spawned by the parent `mokr.browser.Page`.
            ignore_https_errors (bool): Ignore site security errors.
                Inherited from parent `mokr.browser.Page`.
            interception_callback_chain (list[Callable]): A list of callables
                for use with "request" event interception. This list is
                shared by the parent `mokr.browser.Page` and all newly created
                `mokr.network.Request` objects in this manager.

        Returns:
            NetworkManager: New `NetworkManager` with applied configurations.
        """
        network_manager = cls(
            page,
            client,
            frame_manager,
            interception_callback_chain,
        )
        if ignore_https_errors:
            await network_manager._client.send(
                SECURITY_IGNORE_HTTPS_ERRORS,
                {"ignore": True},
            )
        await network_manager._client.send(NETWORK_ENABLE)
        await network_manager._apply_network_conditions()
        await network_manager._apply_protocol_cache_disabled()
        await network_manager._apply_user_agent()
        if not page._is_firefox(mute=True):
            await network_manager._apply_extra_http_headers()
            await network_manager._apply_protocol_request_interception()
        return network_manager

    async def _apply_extra_http_headers(self) -> None:
        if not self._extra_http_headers:
            return
        await self._client.send(
            NETWORK_EXTRA_HEADERS,
            {"headers": self._extra_http_headers},
        )

    async def _apply_network_conditions(self) -> None:
        if not self._emulated_network_conditions:
            return
        await self._client.send(
            NETWORK_EMULATE_NETWORK_CONDITIONS,
            {
                'offline': self._emulated_network_conditions['offline'],
                'latency': self._emulated_network_conditions['latency'],
                'uploadThroughput': self._emulated_network_conditions['upload'],
                'downloadThroughput': (
                    self._emulated_network_conditions['download']
                ),
            }
        )

    async def _apply_protocol_cache_disabled(self) -> None:
        if not self._user_cache_disabled:
            return
        await self._client.send(
            NETWORK_CACHE_DISABLE,
            {'cacheDisabled': self._user_cache_disabled},
        )

    async def _apply_user_agent(self) -> None:
        if not self.user_agent:
            return
        params = {'userAgent': self.user_agent}
        if self.user_agent_metadata:
            params['userAgentMetadata'] = self.user_agent_metadata,
        await self._client.send(NETWORK_USER_AGENT_OVERRIDE, params)

    def _inspect_stack_for_mokr_uuid(self, event: dict) -> str | None:
        mokr_request_uuid = None
        # FetchDomain requests add a unique id to the script URL for tracking.
        stack = event.get("initiator", {}).get("stack", {})
        call_frames = stack.get("callFrames", [])
        if call_frames:
            url = call_frames[0].get("url", "")
            if EVALUATION_SCRIPT_URL in url:
                mokr_request_uuid = url.replace(EVALUATION_SCRIPT_URL, "")
        return mokr_request_uuid

    async def set_extra_http_headers(self) -> None:
        raise NotImplementedError

    async def set_request_interception(self) -> None:
        raise NotImplementedError

    async def set_credentials(self) -> None:
        raise NotImplementedError

    async def emulate_network_conditions(
        self,
        latency: int | None = 0,
        download: int | None = -1,
        upload: int | None = -1,
    ) -> None:
        """
        Emulate the given network conditions. If network conditions unset,
        will set them to the default (disable all throttling and no latency).

        Args:
            latency (int, optional): Minimum latency in milliseconds from
                request sent to response headers received. Defaults to None (0).
            download (int, optional): Maximum download throughput (bytes/sec).
                Defaults to None (-1, disabled).
            upload (int, optional): Minimum download throughput (bytes/sec).
                Defaults to None (-1, disabled).
        """
        if not self._emulated_network_conditions:
            self._emulated_network_conditions = {
                "offline": False,
                "upload": -1,
                "download": -1,
                "latency": 0,
            }
        options = {
            "upload": upload,
            "download": download,
            "latency": latency,
        }
        for option, value in options.items():
            if value is None:
                continue
            self._emulated_network_conditions[option] = value
        await self._apply_network_conditions()

    async def set_offline_mode(self, choice: bool) -> None:
        """
        Enable or diable offline mode. Disabled by default.

        Args:
            enabled (bool): False to disable, True to enable.
        """
        if not self._emulated_network_conditions:
            self._emulated_network_conditions = {
                "offline": False,
                "upload": -1,
                "download": -1,
                "latency": 0,
            }
        self._emulated_network_conditions['offline'] = choice
        await self._apply_network_conditions()

    async def set_request_cache(self, enabled: bool) -> None:
        """
        Enable or disable request caching. Request caching caches requests in
        the browser, not `mokr.network.Request` objects.

        Does not check if request caching was enabled already by parent
        `mokr.browser.Page`.

        Args:
            enabled (bool, optional): True to enable, False to disable.
                Defaults to True.
        """
        self._user_cache_disabled = not enabled
        await self._apply_protocol_cache_disabled()

    async def set_user_agent(
        self,
        user_agent: str,
        user_agent_metadata: str | None = None,
    ) -> None:
        """
        Update the user agent to be sent with every request.

        Args:
            user_agent (str): User agent string.
            user_agent_metadata (str | None, optional): Experimental, used to
                specify user agent client hints to emulate.
        """
        self.user_agent = user_agent
        self.user_agent_metadata = user_agent_metadata
        await self._apply_user_agent()



import logging
from typing import Any, Callable

from pyee import EventEmitter

from mokr.connection import DevtoolsConnection
from mokr.constants import (
    RUNTIME_CONSOLE_API_CALL,
    RUNTIME_ENABLE,
    RUNTIME_EXCEPTION_THROWN,
    RUNTIME_EXECUTION_CONTEXT_CREATED,
)
from mokr.execution.context import ExecutionContext, JavascriptHandle


LOGGER = logging.getLogger(__name__)


class WebWorker(EventEmitter):
    def __init__(
        self,
        client: DevtoolsConnection,
        url: str,
        console_api_callback: Callable,
        exception_thrown: Callable,
    ) -> None:
        """
        Represents a web worker on the page. This object is created and
        destroyed by it's parent `mokr.browser.Page`.

        Args:
            client (DevtoolsConnection): A `mokr.connection.DevtoolsConnection`
                spawned by the parent `mokr.browser.Page`.
            url (str): The "url" present in the triggering event's "targetInfo".
            console_api_callback (Callable): Callback to run when the console
                is called, if no listener's are present for this on the
                `mokr.browser.Page`, will delete the handle due to given
                default callback from parent `Page`.
            exception_thrown (Callable): Callback to run if exception occurs.
        """
        super().__init__()
        self._client = client
        self._url = url
        self.console_api_callback = console_api_callback
        self._execution_context = None
        self._loop = client._loop
        self._execution_context_promise = self._loop.create_future()
        self._client.on(
            RUNTIME_EXECUTION_CONTEXT_CREATED,
            self._on_execution_context_created,
        )
        try:
            # May fail if target closed before all execution contexts received.
            self._client.send(RUNTIME_ENABLE, {})
        except Exception:
            LOGGER.error("Error enabling runtime on webworker.", exc_info=True)
        self._client.on(RUNTIME_CONSOLE_API_CALL, self._on_console_api_called)
        self._client.on(
            RUNTIME_EXCEPTION_THROWN,
            lambda exception: exception_thrown(exception['exceptionDetails']),
        )

    @property
    def url(self) -> str:
        """URL associated with the event that triggered this worker."""
        return self._url

    def _context_js_handle_factory(
        self,
        remote_object: dict,
    ) -> JavascriptHandle | None:
        if self._execution_context:
            return JavascriptHandle(
                self._execution_context,
                self._client,
                remote_object,
            )
        else:
            return None

    def _execution_context_callback(self, value: ExecutionContext) -> None:
        self._execution_context_promise.set_result(value)

    def _on_execution_context_created(self, event: dict) -> None:
        self._execution_context = ExecutionContext(
            self._client,
            event['context'],
            self._context_js_handle_factory,
        )
        self._execution_context_callback(self._execution_context)

    def _on_console_api_called(self, event: dict) -> None:
        args = []
        for arg in event.get('args', []):
            args.append(self._context_js_handle_factory(arg))
        self.console_api_callback(event['type'], args)

    async def execution_context(self) -> ExecutionContext:
        """
        Return the newly created `mokr.execution.ExecutionContext`.

        Returns:
            ExecutionContext: `mokr.execution.ExecutionContext`.
        """
        return await self._execution_context_promise

    async def evaluate(self, page_function: str, *args: Any) -> dict | None:
        """
        Execute a JavaScript function with given arguments.
        Run this `WebWorker`'s `mokr.execution.ExecutionContext.evaluate`.

        Args:
            page_function (str): JavaScript function to run.

        Raises:
            `mokr.exceptions.NetworkError` if an unhandled error occurs
            either evaluating the function or requesting the resulting object.

        Returns:
            dict | None: The decoded object (dict) via
                `mokr.execution.JavascriptHandle.json` or None if a known
                error occurs decoding it.
        """
        return await (await self._execution_context_promise).evaluate(
            page_function,
            *args,
        )

    async def evaluate_handle(
        self,
        page_function: str,
        *args: Any,
    ) -> JavascriptHandle:
        """
        Execute a JavaScript function with given arguments.
        Run this `WebWorker`'s
        `mokr.execution.ExecutionContext.evaluate_handle`.

        Args:
            page_function (str): JavaScript function to run.

        Returns:
            JavascriptHandle: `mokr.execution.JavascriptHandle`.
        """
        return await (await self._execution_context_promise).evaluate_handle(
            page_function,
            *args,
        )



from __future__ import annotations

import asyncio
import base64
import inspect
import json
import logging
import math
import mimetypes
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Literal

from pyee import EventEmitter

from mokr.browser.console import ConsoleMessage
from mokr.browser.viewport import ViewportManager
from mokr.browser.worker import WebWorker
from mokr.connection import DevtoolsConnection
from mokr.constants import (
    CLOSE,
    DOM_LOADED,
    EMULATION_DISABLE_SCRIPT_EXECUTION,
    EMULATION_OVERRIDE_BACKGROUND,
    EMULATION_OVERRIDE_METRICS,
    EMULATION_SET_EMULATED_MEDIA,
    ERROR,
    FRAME_ATTACHED,
    FRAME_DETACHED,
    FRAME_NAVIGATED,
    HTTP_METHODS,
    INSPECTOR_TARGET_CRASHED,
    LIFECYCLE_EVENTS,
    LOG_ENABLE,
    LOG_ENTRY_ADDED,
    METHOD_ADD_PAGE_BINDING,
    METHOD_DELIVER_BINDING_RESULT,
    METRICS,
    NETWORK_CACHE_DISABLE,
    NETWORK_DELETE_COOKIES,
    NETWORK_ENABLE,
    NETWORK_GET_ALL_COOKIES,
    NETWORK_GET_COOKIES,
    NETWORK_MGR_REQUEST_FINISHED,
    NETWORK_MGR_RESPONSE,
    NETWORK_REQUEST_FAILED,
    NETWORK_REQUEST_FINISHED,
    NETWORK_RESPONSE,
    NETWORK_SET_COOKIES,
    PAGE_ADD_SCRIPT_TO_EVAL,
    PAGE_BRING_TO_FRONT,
    PAGE_CLOSE,
    PAGE_CONSOLE,
    PAGE_DIALOG,
    PAGE_DOM_LOADED,
    PAGE_ENABLE,
    PAGE_ENABLE_LIFECYCLE_EVENTS,
    PAGE_ERROR,
    PAGE_GET_FRAME_TREE,
    PAGE_GET_LAYOUT,
    PAGE_GET_NAVIGATION_HISTORY,
    PAGE_JAVASCRIPT_DIALOG_OPEN,
    PAGE_LOAD,
    PAGE_LOAD_EVENT_FIRED,
    PAGE_NAVIGATE,
    PAGE_NAVIGATE_TO_HISTORY_ENTRY,
    PAGE_PRINT_TO_PDF,
    PAGE_SCREENSHOT,
    PAGE_SET_BYPASS_CSP,
    PERFORMANCE_ENABLE,
    PERFORMANCE_GET_METRICS,
    PERFORMANCE_METRICS,
    RUMTIME_BINDING_CALL,
    RUNTIME_ADD_BINDING,
    RUNTIME_CONSOLE_API_CALL,
    RUNTIME_ENABLE,
    RUNTIME_EVALUATE,
    RUNTIME_EXCEPTION_THROWN,
    SECURITY_ENABLE,
    TARGET_ACTIVATE,
    TARGET_ATTACHED,
    TARGET_CLOSE,
    TARGET_DETACHED,
    TARGET_SEND_DETACH,
    TARGET_SET_AUTO_ATTACH,
    WORKER_CREATED,
    WORKER_DESTROYED,
    NETWORK_MGR_REQUEST,
    PAGE_RELOAD
)
from mokr.exceptions import FirefoxNotImplementedError, PageError
from mokr.execution import ElementHandle, JavascriptHandle
from mokr.frame import Frame, FrameManager
from mokr.input import Dialog, Keyboard, Mouse, Touchscreen
from mokr.network import (
    ChromeNetworkManager,
    FetchDomain,
    FirefoxNetworkManager,
    HttpDomain,
    NetworkManager,
    Request,
    Response,
)
from mokr.utils.remote import (
    add_event_listener,
    format_javascript_exception,
    release_remote_object,
    remove_event_listeners,
    serialize_remote_object,
)
from mokr.waiters import EventWaiter, NavigationWaiter

if TYPE_CHECKING:
    from mokr.browser import Browser
    from mokr.browser.target import Target


LOGGER = logging.getLogger(__name__)


class Page(EventEmitter):
    def __init__(
        self,
        browser: Browser,
        client: DevtoolsConnection,
        target: Target,
        frame_tree: dict,
        ignore_https_errors: bool,
        user_agent: str = None,
        screenshot_task_queue: list = None,
        proxy_credentials: dict | None = None,
    ) -> None:
        """
        Representative of a single tab within the browser (more specifically,
        within the browser context).
        This class is laregly the entry-point for most navigation, execution,
        or dom-manipulation tasks. Many methods within this class are simply
        wrappers to bound classes within it.

        `Page` emits events, often "inherited" from other classes (emitted
        in series on recieving them from bound classes); these events can be
        listened on via `Page.on`.
        Note, however, that certain network events will be bound directly to
        this `Page`'s `mokr.network.NetworkManager` when passed to `Page.on`.
        Additionally, `Page.on("request")` will not behaviour as a regular
        listener, but instead be added to the request interception callback
        chain. See `Page.on` for more.

        Args:
            browser (Browser): Parent `mokr.browser.Browser`.
            client (DevtoolsConnection): From parent `mokr.browser.Target`.
            target (Target): Spawning `mokr.browser.Target`.
            frame_tree (dict): Frame tree structure from remote connection.
            ignore_https_errors (bool): Ignore site security errors.
                Inherited from parent `mokr.browser.Browser`.
            user_agent (str): User agent for this page.
                Inherited from parent `mokr.browser.Browser`.
            screenshot_task_queue (list, optional): The screenshot task queue,
                empty by default. Inherited from parent `mokr.browser.Browser`.
            proxy_credentials: (dict | None, optional): Dictionary with proxy
                credentials keyed as "username" and "password". Credentials
                should be for the proxy the browser process is bound to.
                Inherited from parent `mokr.browser.Browser`.
        """
        super().__init__()
        self._browser = browser
        self._closed = False
        self._client = client
        self._target = target
        self._keyboard = Keyboard(client)
        self._mouse = Mouse(client, self._keyboard)
        self._touchscreen = Touchscreen(client, self._keyboard)
        self._frame_manager = FrameManager(client, frame_tree, self)
        self._network_manager: NetworkManager = None
        self._viewport_manager = ViewportManager(client)
        self._page_bindings: dict[str, Callable] = {}
        self._ignore_https_errors = ignore_https_errors
        self._default_navigation_timeout = 30000
        self._javascript_enabled = True
        self._viewport: dict | None = None
        self._proxy_credentials = proxy_credentials if proxy_credentials else {}
        self._user_agent = user_agent
        if screenshot_task_queue is None:
            screenshot_task_queue = []
        self._screenshot_task_queue = screenshot_task_queue
        self._workers: dict[str, WebWorker] = {}
        client.on(TARGET_ATTACHED, self._on_target_attached)
        client.on(TARGET_DETACHED, self._on_target_detached)
        for event_name in [FRAME_ATTACHED, FRAME_DETACHED, FRAME_NAVIGATED]:
            self._frame_manager.on(
                event_name, lambda event: self.emit(event_name, event)
            )
        client_events_to_methods = {
            PAGE_DOM_LOADED: lambda event: self.emit(DOM_LOADED),
            PAGE_LOAD_EVENT_FIRED: lambda event: self.emit(PAGE_LOAD),
            RUNTIME_CONSOLE_API_CALL: lambda event: self._on_console_api(event),
            RUMTIME_BINDING_CALL: lambda event: self._on_binding_called(event),
            PAGE_JAVASCRIPT_DIALOG_OPEN: lambda event: self._on_dialog(event),
            RUNTIME_EXCEPTION_THROWN: lambda exception: self._handle_exception(
                exception.get('exceptionDetails')
            ),
            INSPECTOR_TARGET_CRASHED: lambda event: self._on_target_crashed(),
            PERFORMANCE_METRICS: lambda event: self._emit_metrics(event),
            LOG_ENTRY_ADDED: lambda event: self._on_log_entry_added(event),
        }
        for event_name, method in client_events_to_methods.items():
            client.on(event_name, method)
        self._target._is_closed_promise.add_done_callback(self._set_closed)
        self._interception_callback_chain = [self._default_request_intercept]
        self._callback_chain_listeners = {}
        self._fetch_domain = None
        self._http_domain = None

    @staticmethod
    async def create(
        browser: Browser,
        client: DevtoolsConnection,
        target: Target,
        ignore_https_errors: bool,
        default_viewport: dict | None,
        screenshot_task_queue: list = None,
        proxy_credentials: dict | None = None,
        user_agent_data: dict[str, str] | None = None,
    ) -> Page:
        """
        Async constructor for this class. Necessary to run some asyncronous
        post-initialisation tasks.

        Args:
            browser (Browser): Parent `mokr.browser.Browser`.
            client (DevtoolsConnection): From parent `mokr.browser.Target`.
            target (Target): Spawning `mokr.browser.Target`.
            ignore_https_errors (bool): Ignore site security errors.
                Inherited from parent `mokr.browser.Browser`.
            default_viewport (dict | None): Default viewport configuration.
            screenshot_task_queue (list, optional): The screenshot task queue,
                empty by default. Inherited from parent `mokr.browser.Browser`.
            proxy_credentials: (dict | None, optional): Dictionary with proxy
                credentials keyed as "username" and "password". Credentials
                should be for the proxy the browser process is bound to.
                Inherited from parent `mokr.browser.Browser`.
            user_agent_data (dict[str, str] | None, optional): A dictionary
                containing the user agent and an indicator to whether it was the
                original or an override.

        Returns:
            Page: New `Page` with necessary remote configuration enabled.
        """
        await client.send(PAGE_ENABLE)
        frame_tree = (await client.send(PAGE_GET_FRAME_TREE))['frameTree']
        user_agent = user_agent_data["user_agent"]
        page = Page(
            browser,
            client,
            target,
            frame_tree,
            ignore_https_errors,
            user_agent,
            screenshot_task_queue,
            proxy_credentials,
        )
        network_manager_classes = {
            "chrome": ChromeNetworkManager,
            "firefox": FirefoxNetworkManager,
        }
        network_manager_cls = network_manager_classes.get(browser.kind)
        network_manager: NetworkManager = await network_manager_cls.create(
            page,
            page._client,
            page._frame_manager,
            page._ignore_https_errors,
            page._interception_callback_chain,
        )
        if page._is_firefox(mute=True):
            # Set this to true manually because it cannot be unset and while
            # interception is lacking, it is technically always on.
            network_manager._user_request_interception_enabled = True
        page._network_manager = network_manager
        # TODO emitting a "response" on page from network_mgr_response doesn't
        # behave as expected. Needs further investigaiton, for now we'll alias
        # this in page.on.
        for network_event in [
            NETWORK_MGR_REQUEST,
            NETWORK_MGR_RESPONSE,
            NETWORK_REQUEST_FAILED,
            NETWORK_MGR_REQUEST_FINISHED,
        ]:
            network_manager.on(
                network_event,
                lambda event: page.emit(network_event, event),
            )
        await asyncio.gather(
            client.send(
                TARGET_SET_AUTO_ATTACH,
                {'autoAttach': True, 'waitForDebuggerOnStart': False},
            ),
            client.send(PAGE_ENABLE_LIFECYCLE_EVENTS, {'enabled': True}),
            client.send(NETWORK_ENABLE, {}),
            client.send(RUNTIME_ENABLE, {}),
            client.send(SECURITY_ENABLE, {}),
            client.send(PERFORMANCE_ENABLE, {}),
            client.send(LOG_ENABLE, {}),
        )
        is_firefox = page._is_firefox(mute=True)
        if default_viewport:
            await page.set_viewport(default_viewport)
        if not is_firefox:
            await page.set_request_interception_enabled(True)
        if user_agent_data and user_agent_data.get("overridden"):
            await page.set_user_agent(user_agent)
        if proxy_credentials and not is_firefox:
            await page.set_credentials(proxy_credentials)
        page._fetch_domain = FetchDomain(page)
        return page

    @staticmethod
    def _make_javascript_function_string(method_name: str, *args: Any) -> str:
        # Convert function and arguments to str.
        _args = ', '.join(
            [
                json.dumps('undefined' if arg is None else arg) for arg in args
            ]
        )
        expr = f'({method_name})({_args})'
        return expr

    @staticmethod
    async def _evaluate(frame: Frame, expression: str) -> None:
        try:
            await frame.evaluate(expression, force_expr=True)
        except Exception:
            LOGGER.error(
                f"Error evaluating expression in frame: {expression}",
                exc_info=True,
            )

    @staticmethod
    def _stash_response(
        all_responses: dict[str, Response],
        request: Request,
    ) -> None:
        responses = all_responses.get(request.url, [])
        if request.response:
            responses.append(request.response)
        all_responses[request.url] = responses

    @staticmethod
    def _convert_print_param(
        parameter: None | int | float | str
    ) -> float | None:
        # Convert print parameter to inches.
        unit_to_pixels = {'px': 1, 'in': 96, 'cm': 37.8, 'mm': 3.78}
        if parameter is None:
            return None
        if isinstance(parameter, (int, float)):
            pixels = parameter
        elif isinstance(parameter, str):
            text = parameter
            unit = text[-2:].lower()
            if unit in unit_to_pixels:
                value_text = text[:-2]
            else:
                unit = 'px'
                value_text = text
            try:
                value = float(value_text)
            except ValueError:
                raise ValueError(f'Failed to parse parameter value: {text}')
            pixels = value * unit_to_pixels[unit]
        else:
            raise TypeError(f'Cannot accept type: {str(type(parameter))}')
        return pixels / 96

    @property
    def target(self) -> Target:
        """Parent `mokr.browser.Target` this `Page` was spawned from."""
        return self._target

    @property
    def browser(self) -> Browser:
        """Parent `mokr.browser.Browser` this `Page` belongs to."""
        return self._target.browser

    @property
    def main_frame(self) -> Frame | None:
        """This `Page`'s `mokr.frame.FrameManager`'s main `mokr.frame.Frame`."""
        return self._frame_manager._main_frame

    @property
    def network_manager(self) -> NetworkManager | None:
        """This `Page`'s `mokr.network.NetworkManager`."""
        return self._network_manager

    @property
    def fetch_domain(self) -> FetchDomain:
        """This `Page`'s `mokr.fetch.FetchDomain`."""
        return self._fetch_domain

    @property
    def http_domain(self) -> HttpDomain:
        """This `Page`'s `mokr.network.HttpDomain`."""
        if not self._http_domain:
            self.make_http_domain()
        return self._http_domain

    @property
    def keyboard(self) -> Keyboard:
        """This `Page`'s `mokr.input.Keyboard`."""
        return self._keyboard

    @property
    def mouse(self) -> Mouse:
        """This `Page`'s `mokr.input.Mouse`."""
        return self._mouse

    @property
    def touchscreen(self) -> Touchscreen:
        """This `Page`'s `mokr.input.Touchscreen`."""
        return self._touchscreen

    @property
    def frames(self) -> list[Frame]:
        """List of all `mokr.frame.Frames` within this `Page`."""
        return list(self._frame_manager.frames())

    @property
    def workers(self) -> list[WebWorker]:
        """List of all `mokr.browser.WebWorker`s within this `Page`."""
        return list(self._workers.values())

    @property
    def url(self) -> str:
        """The URL the `Page` is currently at."""
        frame = self._ensure_frame()
        return frame.url

    @property
    def viewport(self) -> dict | None:
        """Current viewport configuration, if any."""
        return self._viewport

    @property
    def user_agent(self) -> str | None:
        """The user agent given to override on all requests, if any."""
        return self._user_agent

    @property
    def is_closed(self) -> bool:
        """
        True if the `Page` is still alive, otherwise False.
        If a parent `mokr.browser.BrowserContext` is closed, all child `Page`s
        will close, too.
        """
        return self._closed

    @property
    def browser_type(self) -> str:
        """One of "chrome" or "firefox"."""
        return self._browser.kind

    def _is_firefox(self, mute: bool = False, caller: Any = None) -> bool:
        if not self.browser_type == "firefox":
            return False
        elif mute:
            return True
        else:
            name = inspect.currentframe().f_back.f_code.co_name
            if caller:
                name = f"{caller.__class__.__name__}.{name}"
            raise FirefoxNotImplementedError(
                f"Firefox doesn't support functionality to use method {name}."
            )

    def _set_closed(self, *args) -> None:
        self.emit(CLOSE)
        self._closed = True

    def _on_target_attached(self, event: dict) -> None:
        target_info = event['targetInfo']
        if target_info['type'] != 'worker':
            # If we don't detach from service workers, they will never die.
            try:
                self._client.send(
                    TARGET_SEND_DETACH,
                    {'sessionId': event['sessionId']},
                )
            except Exception:
                LOGGER.info("Error detaching worker.", exc_info=True)
            return
        session_id = event['sessionId']
        session = self._client._create_session(target_info['type'], session_id)
        worker = WebWorker(
            session,
            target_info['url'],
            self._add_console_message,
            self._handle_exception,
        )
        self._workers[session_id] = worker
        self.emit(WORKER_CREATED, worker)

    def _on_target_detached(self, event: dict) -> None:
        session_id = event['sessionId']
        worker = self._workers.get(session_id)
        if worker is None:
            return
        self.emit(WORKER_DESTROYED, worker)
        self._workers.pop(session_id)

    def _on_target_crashed(self, *args: Any, **kwargs: Any) -> None:
        self.emit(ERROR, PageError('Page crashed!'))

    def _on_log_entry_added(self, event: dict) -> None:
        entry = event.get('entry', {})
        level = entry.get('level', '')
        text = entry.get('text', '')
        args = entry.get('args', [])
        source = entry.get('source', '')
        for arg in args:
            release_remote_object(self._client, arg)
        if source != 'worker':
            self.emit(PAGE_CONSOLE, ConsoleMessage(level, text))

    def _ensure_frame(self) -> Frame:
        frame = self.main_frame
        if not frame:
            raise PageError('Page has no main frame.')
        return frame

    def _emit_metrics(self, event: dict) -> None:
        self.emit(
            METRICS,
            {
                'title': event['title'],
                'metrics': self._build_metrics(event.get("metrics", [])),
            }
        )

    def _build_metrics(self, metrics: list) -> dict[str, Any]:
        supported_metrics = (
            'Timestamp',
            'Documents',
            'Frames',
            'JSEventListeners',
            'Nodes',
            'LayoutCount',
            'RecalcStyleCount',
            'LayoutDuration',
            'RecalcStyleDuration',
            'ScriptDuration',
            'TaskDuration',
            'JSHeapUsedSize',
            'JSHeapTotalSize',
        )
        result = {}
        for metric in metrics:
            if metric['name'] in supported_metrics:
                result[metric['name']] = metric['value']
        return result

    def _handle_exception(self, exceptionDetails: dict) -> None:
        message = format_javascript_exception(exceptionDetails)
        self.emit(PAGE_ERROR, PageError(message))

    def _on_console_api(self, event: dict) -> None:
        context = self._frame_manager.execution_context_by_id(
            event['executionContextId']
        )
        values = []
        for arg in event.get('args', []):
            values.append(
                self._frame_manager.create_javascript_handle(context, arg)
            )
        self._add_console_message(event['type'], values)

    def _on_binding_called(self, event: dict) -> None:
        obj = json.loads(event['payload'])
        name = obj['name']
        seq = obj['seq']
        args = obj['args']
        result = self._page_bindings[name](*args)
        expression = self._make_javascript_function_string(
            METHOD_DELIVER_BINDING_RESULT,
            name,
            seq,
            result,
        )
        try:
            self._client.send(
                RUNTIME_EVALUATE,
                {
                    'expression': expression,
                    'context_id': event['executionContextId'],
                },
            )
        except Exception:
            LOGGER.error(
                f"Error evaluating expression: {expression}",
                exc_info=True,
            )

    def _add_console_message(
        self,
        type: str,
        args: list[JavascriptHandle],
    ) -> None:
        if not self.listeners(PAGE_CONSOLE):
            for arg in args:
                self._client._loop.create_task(arg.dispose())
            return
        text_tokens = []
        for arg in args:
            remote_object = arg._remote_object
            if remote_object.get('objectId'):
                text_tokens.append(arg.to_string())
            else:
                text_tokens.append(str(serialize_remote_object(remote_object)))
        message = ConsoleMessage(type, ' '.join(text_tokens), args)
        self.emit(PAGE_CONSOLE, message)

    def _on_dialog(self, event: Any) -> None:
        _type = event.get('type')
        if _type == 'alert':
            dialog_type = _type
        elif _type == 'confirm':
            dialog_type = _type
        elif _type == 'prompt':
            dialog_type = _type
        elif _type == 'beforeunload':
            dialog_type = _type
        else:
            dialog_type = ''
        if self._is_firefox(mute=True):
            dialog = event
        else:
            dialog = Dialog(
                self._client,
                dialog_type,
                event.get('message'),
                event.get('defaultPrompt'),
            )
        self.emit(PAGE_DIALOG, dialog)

    async def _default_request_intercept(self, request: Request) -> None:
        # Default interception is to just not block the request.
        if not self._is_firefox(mute=True):
            return await request.release()

    async def _navigate(self, url: str, referrer: str) -> str | None:
        response = await self._client.send(
            PAGE_NAVIGATE,
            {'url': url, 'referrer': referrer},
        )
        if response.get('errorText'):
            return f'{response["errorText"]} at {url}'
        return None

    async def _screenshot_task(
        self,
        file_type: Literal["png", "jpeg"] | None,
        file_path: str | None,
        jpeg_quality: int | None,
        full_page: bool,
        clip: dict[str, int] | None,
        omit_background: bool,
        encoding: Literal["binary", "base64"],
        scale: int,
    ) -> bytes:
        await self._client.send(
            TARGET_ACTIVATE,
            {'targetId': self._target._targetId},
        )
        if clip:
            clip['scale'] = 1
        if full_page:
            metrics = await self._client.send(PAGE_GET_LAYOUT)
            width = math.ceil(metrics['contentSize']['width'])
            height = math.ceil(metrics['contentSize']['height'])
            # Overwrite clip for full page.
            clip = dict(x=0, y=0, width=width, height=height, scale=1)
            if self._viewport is not None:
                mobile = self._viewport.get('isMobile', False)
                device_scale_factor = self._viewport.get(
                    'deviceScaleFactor',
                    scale,
                )
                landscape = self._viewport.get('isLandscape', False)
            else:
                mobile = False
                device_scale_factor = scale
                landscape = False
            if landscape:
                screen_orientation = dict(angle=90, type='landscapePrimary')
            else:
                screen_orientation = dict(angle=0, type='portraitPrimary')
            await self._client.send(
                EMULATION_OVERRIDE_METRICS,
                {
                    'mobile': mobile,
                    'width': width,
                    'height': height,
                    'deviceScaleFactor': device_scale_factor,
                    'screenOrientation': screen_orientation,
                },
            )
        if omit_background and not self._is_firefox(mute=True):
            await self._client.send(
                EMULATION_OVERRIDE_BACKGROUND,
                {'color': {'r': 0, 'g': 0, 'b': 0, 'a': 0}},
            )
        opt = {'format': file_type}
        if clip:
            opt['clip'] = clip
        if file_type == 'jpeg' and jpeg_quality is not None:
            opt['quality'] = jpeg_quality
        # Send the screenshot request with the given parameters.
        response = await self._client.send(PAGE_SCREENSHOT, opt)
        # Restore any overrides.
        if omit_background and not self._is_firefox(mute=True):
            await self._client.send(EMULATION_OVERRIDE_BACKGROUND)
        if full_page and self._viewport is not None:
            await self.set_viewport(self._viewport)
        # Encode and write file, if requested.
        if encoding == 'base64':
            buffer = response.get('data', b'')
        else:
            buffer = base64.b64decode(response.get('data', b''))
        if file_path:
            with open(file_path, 'wb') as f:
                f.write(buffer)
        return buffer

    async def _navigate_history(
        self,
        delta: int,
        timeout: int | None = None,
        wait_until: list[LIFECYCLE_EVENTS] | LIFECYCLE_EVENTS = "load",
    ) -> Response | None:
        history = await self._client.send(PAGE_GET_NAVIGATION_HISTORY)
        _count = history.get('currentIndex', 0) + delta
        entries = history.get('entries', [])
        if len(entries) <= _count:
            return None
        entry = entries.get(_count, {})
        response = (
            await asyncio.gather(
                self.wait_for_navigation(timeout, wait_until),
                self._client.send(
                    PAGE_NAVIGATE_TO_HISTORY_ENTRY,
                    {'entryId': entry.get('id')}
                ),
            )
        )[0]
        return response

    def on(self, event: str, method: Callable | None) -> None:
        """
        Registers the callable `method` to the event name `event`, if provided.
        When the target `event` is emitted by this class, the target `method`
        will be called.

        Notably, the event "request" does not register a listener. Instead, it
        adds the method to this `Page`'s `mokr.network.NetworkManager`'s
        request interception callback chain, which is already registered
        as a listener on "request" events, if interception is enabled.

        This callback chain allows multiple manipulations of requests, until the
        intercepted `mokr.network.Request` is `Request.released`, `.aborted`, or
        `.fulfilled`. For request interception, methods are executed in reverse
        order of registration, so newer `Page.on` methods will run first.

        Args:
            event (str): Event name to listen for.
            method (Callable | None): Method to call when event is emitted.

        Example::

            # Printing worker URL on creation, standard functionality.
            page.on("workercreated", lambda worker: print(worker.url))

            # Adding a handler to the request interception chain.
            async def intercept(request: Request) -> Request | None:
                if request.url.endswith("png"):
                    # Request will be killed, interception chain stopped here.
                    await request.abort()
                elif request.url.endswith("jpg"):
                    # Request will continue in the interception chain.
                    return request
                else:
                    # Request will be finished, interception chain stopped here.
                    await request.release()
            # Handlers can be synchronous, too.
            def note_url(request: Request) -> Request:
                print(request.url)
                return request
            page.on("request", note_url)
            page.on("request", intercept)
        """
        # Request interception may act strangely with stylesheet requests.
        if event == "request":
            if method in self._interception_callback_chain:
                self._interception_callback_chain.remove(method)
            self._interception_callback_chain.insert(0, method)
        else:
            _network_manager_on = False
            network_mgr_events = {
                NETWORK_RESPONSE: NETWORK_MGR_RESPONSE,
                NETWORK_REQUEST_FINISHED: NETWORK_MGR_REQUEST_FINISHED,
            }
            if event in network_mgr_events.keys():
                event = network_mgr_events.get(event)
                _network_manager_on = True
            if inspect.iscoroutinefunction(method):
                callback = lambda *args, **kwargs: asyncio.ensure_future(
                    method(*args, **kwargs)
                )
            else:
                callback = method
            _cls = self._network_manager if _network_manager_on else super()
            _cls.on(event, callback)

    def make_http_domain(
        self,
        sync_cookies: Literal["both", "http", "page", "none"] = "both",
        **httpx_client_kwargs,
    ) -> None:
        """
        Create a `mokr.network.HttpDomain` bound to this page.

        This method does not ever need to be called directly; accessing
        `Page.http_domain` will create a new domain with default values. This
        exists so any optional arguments could be passed to the constructor.

        Optionally set the cookie sync behaviour by setting `sync_cookies` to:
        - "both" (default): Sync back and forth.
        - "http": Sync to the `HttpDomain` from the `Page` only.
        - "page": Sync to the `Page` from the `HttpDomain` only.
        - "none": Don't sync at all.

        Args:
            sync_cookies (Literal["both", "http", "page", "none"], optional):
                Whether to sync cookies between the `HttpDomain` and the parent
                `Page` before/after each request and response.
                Defaults to "both".
        """
        self._http_domain = HttpDomain(
            self,
            sync_cookies,
            **httpx_client_kwargs,
        )

    async def set_request_interception_enabled(self, choice: bool) -> None:
        """
        Disable or enabled request interception (enabled by default).
        See `Page.on` for details on using the request interception
        callback chain.

        Args:
            choice (bool): False to disable, True to enable.
        """
        await self._network_manager.set_request_interception(choice)

    async def tap(self, selector: str) -> None:
        """
        Tap the first element that matches `selector`.
        Wrapper for `Page.main_frame.tap`.

        This method is a shortcut for running `Page.query_selector` and then
        running `mokr.execution.ElementHandle.tap` on the resultant
        ElementHandle. In either case, an element will be scrolled into view
        if needed, and the center of it clicked with the ElementHandle's
        bound `Page.touchscreen`.

        Args:
            selector (str): Selector to query element by.

        Raises:
            PageError: Raised if no element is found with given `selector`.
        """
        self._is_firefox(caller=self)
        frame = self._ensure_frame()
        await frame.tap(selector)

    async def set_offline_mode(self, choice: bool) -> None:
        """
        Enable or diable offline mode. Disabled by default.
        Wrapper for `Page.network_manager.set_offline_mode`.

        Args:
            enabled (bool): False to disable, True to enable.
        """
        await self._network_manager.set_offline_mode(choice)

    def set_default_navigation_timeout(self, timeout: int) -> None:
        """
        Change the default navigation timeout. This is used by `Page.goto`,
        `Page.go_back`, `Page.go_forward`, `Page.wait_for_navigation`,
        `Page.refresh`, and `Page.fetch`, unless overridden.

        Args:
            timeout (int): Milliseconds to wait. Use 0 for infinite.
        """
        self._default_navigation_timeout = timeout

    async def evaluate_handle(
        self,
        page_function: str,
        *args: Any,
    ) -> JavascriptHandle:
        """
        Execute a JavaScript function with given arguments.
        Runs this `Page`'s default `mokr.frame.Frame.evaluate_handle`.

        Args:
            page_function (str): JavaScript function to run.

        Raises:
            PageError: Raised if the `Page`'s main `mokr.frame.Frame` fails to
                find its `mokr.browser.ExecutionContext`.

        Returns:
            JavascriptHandle: `mokr.execution.JavascriptHandle`.
        """
        frame = self._ensure_frame()
        context = await frame.execution_context()
        if not context:
            raise PageError('No context attached to frame.')
        return await context.evaluate_handle(page_function, *args)

    async def query_objects(
        self,
        javascript_handle: JavascriptHandle,
    ) -> JavascriptHandle:
        """
        Query all objects with the given `mokr.execution.JavascriptHandle`'s
        `get_property("objectId")`.
        Wrapper for this `Page.main_frame`'s
        `mokr.execution.ExecutionContext.query_objects`.

        Args:
            javascript_handle (JavascriptHandle): A valid (not disposed)
                `mokr.execution.JavascriptHandle` object.

        Raises:
            PageError: Raised if the `Page`'s main `mokr.frame.Frame` fails to
                find its `mokr.browser.ExecutionContext`.
            ElementHandleError: Raised if the given handle is disposed, or
                does not have an "objectId" remote property (primitive type).

        Returns:
            JavascriptHandle: A `mokr.execution.JavascriptHandle` initialised
                from the remot response.
        """
        self._is_firefox(caller=self)
        frame = self._ensure_frame()
        context = await frame.execution_context()
        if not context:
            raise PageError('No context attached to frame.')
        return await context.query_objects(javascript_handle)

    async def query_selector(self, selector: str) -> ElementHandle | None:
        """
        Return the first element in the DOM that matches the selector, if any.
        Wrapper for `Page.main_frame.query_selector`.

        Args:
            selector (str): Element selector to locate.

        Returns:
            ElementHandle | None: ElementHandle if found or None.
        """
        frame = self._ensure_frame()
        return await frame.query_selector(selector)

    async def query_selector_all(self, selector: str) -> list[ElementHandle]:
        """
        Return all elements in the DOM that match the selector, if any.
        Wrapper for `Page.main_frame.query_selector_all`.

        Args:
            selector (str): Element selector to locate.

        Returns:
            list[ElementHandle]: List of ElementHandle if any or empty list.
        """
        frame = self._ensure_frame()
        return await frame.query_selector_all(selector)

    async def xpath(self, expression: str) -> list[ElementHandle]:
        """
        Return all elements in the DOM that match the expression, if any.
        Wrapper for `Page.main_frame.xpath`.

        Args:
            expression (str): XPath expression to evaluate.

        Returns:
            list[ElementHandle]: List of ElementHandle if any or empty list.
        """
        frame = self._ensure_frame()
        return await frame.xpath(expression)

    async def cookies(self) -> list[dict[str, str | int | bool]]:
        """
        Get all cookies accessible to the browser.

        Returns:
            list[dict[str, str | int | bool]]: List of dictionaries representing
                each cookie.
        """
        response = await self._client.send(NETWORK_GET_ALL_COOKIES)
        return response.get('cookies', [])

    async def get_cookies_by_urls(
        self,
        urls: list[str] | None = None,
    ) -> list[dict[str, str | int | bool]]:
        """
        Get cookies for the given URLs. If no URLs are given, defaults to a
        list containing the URLs of the page and all of its frames.

        Args:
            urls (list[str] | None): List of URLs to get cookies for.
                Defaults to None.

        Returns:
            list[dict[str, str | int | bool]]: List of dictionaries representing
                each cookie.
        """
        params = {}
        if urls:
            urls = params["urls"] = tuple(urls)
        response = await self._client.send(NETWORK_GET_COOKIES, params)
        return response.get('cookies', [])

    async def delete_cookies(self, cookies: list[dict]) -> None:
        """
        Delete a cookie from this `Page`.

        Args:
            cookies (list[dict]): A list of dictionaries representing a cookie
                entry. Each dictionary must at minimum contain a "name".
        """
        for cookie in cookies:
            await self._client.send(NETWORK_DELETE_COOKIES, cookie)

    async def set_cookies(self, cookies: list[dict]) -> None:
        """
        Set a cookie on this `Page`.

        Args:
            cookies (list[dict]): A list of dictionaries representing a cookie
                entry. Each dictionary must at minimum contain a "name" and
                a "value".

        Raises:
            PageError: Raised if the current page is blank ("about:blank") or
                is a data URL (starts with "data:").
        """
        items = []
        for cookie in cookies:
            item = cookie.copy()
            if item.get('url') == 'about:blank':
                name = item.get('name', '')
                raise PageError(f'Blank page can not have cookie "{name}"')
            if item.get('url', '').startswith('data:'):
                name = item.get('name', '')
                raise PageError(f'Data URL page can not have cookie "{name}"')
            items.append(item)
        await self.delete_cookies(items)
        if items:
            await self._client.send(NETWORK_SET_COOKIES, {'cookies': items})

    async def embed_javascript(
        self,
        file_content: str = None,
        file_path: str = None,
        url: str = None,
        script_type: str = None,
    ) -> ElementHandle:
        """
        Add script tag to this `Page.main_frame`.
        Wrapper for `Page.main_frame.embed_javascript`.

        Args:
            file_content (str, optional): Encoded file content for script to
                embed. If not given, must give `file_path` or `url`. Defaults
                to None.
            file_path (str, optional): File path for script to embed.
                If not given, must give `file_content` or `url`. Defaults to
                None.
            url (str, optional): URL for script to embed. If not given, must
                give `file_content` or `file_path`. Defaults to None.
            script_type (str, optional): Use "module" to load as JavaScript
                ES6 module, if not given, defaults to "text/javascript".
                Defaults to None.

        Raises:
            ValueError: Raised if none of `file_content`, `file_path`, or
                `url` are given.
            PageError: Raised if error occurs sending embed request to remote
                connection (from ElementHandleError).

        Returns:
            ElementHandle: Newly embedded element.
        """
        frame = self._ensure_frame()
        return await frame.embed_javascript(
            file_content,
            file_path,
            url,
            script_type,
        )

    async def embed_style(
        self,
        file_content: str = None,
        file_path: str = None,
        url: str = None,
    ) -> ElementHandle:
        """
        Add style tag to this `Page.main_frame`.
        Wrapper for `Page.main_frame.embed_style`.

        Args:
            file_content (str, optional): Encoded file content for style to
                embed. If not given, must give `file_path` or `url`. Defaults
                to None.
            file_path (str, optional): File path for style to embed.
                If not given, must give `file_content` or `url`. Defaults to
                None.
            url (str, optional): URL for style to embed. If not given, must
                give `file_content` or `file_path`. Defaults to None.

        Raises:
            ValueError: Raised if none of `file_content`, `file_path`, or
                `url` are given.
            PageError: Raised if error occurs sending embed request to remote
                connection (from ElementHandleError).

        Returns:
            ElementHandle: Newly embedded element.
        """
        frame = self._ensure_frame()
        return await frame.embed_style(file_content, file_path, url)

    async def expose_function(
        self,
        name: str,
        method: Callable
    ) -> None:
        """
        Bind a remote JavaScript function "name" to local `method`. This will
        allow triggering local `method` by calling `window["name"]()` in the
        browser.

        Args:
            name (str): Remote JavaScript function name.
            method (Callable): Local callable to bind to.

        Raises:
            PageError: Raised if a function exists already with given "name".
        """
        self._is_firefox(caller=self)
        if self._page_bindings.get(name):
            raise PageError(
                f'Failed to add page binding with name {name}:'
                f' window["{name}"] already exists.'
            )
        self._page_bindings[name] = method
        expression = self._make_javascript_function_string(
            METHOD_ADD_PAGE_BINDING,
            name,
        )
        await self._client.send(RUNTIME_ADD_BINDING, {'name': name})
        await self._client.send(PAGE_ADD_SCRIPT_TO_EVAL, {'source': expression})
        loop = asyncio.get_event_loop()
        await asyncio.wait(
            [
                loop.create_task(self._evaluate(frame, expression))
                for frame in self.frames
            ]
        )

    async def set_credentials(
        self,
        credentials: dict[Literal["username", "password"], str],
    ) -> None:
        """
        Set the credentials to use for HTTP authentication.
        Wrapper for `Page.network_manager.set_credentials`.

        Args:
            credentials (dict[str, str]): A dictionary with credentials,
                keyed as "username" and "password".
        """
        self._is_firefox(caller=self)
        credentials = credentials.copy()
        if credentials.copy().get("proxy"):
            credentials.pop("proxy")
        if credentials:
            return await self._network_manager.set_credentials(credentials)

    async def set_extra_http_headers(self, headers: dict[str, str]) -> None:
        """
        Set extra headers to be sent with every request.
        Wrapper for `Page.network_manager.set_extra_http_headers`.

        Args:
            extra_http_headers (dict[str, str]): A dictionary of headers.

        Raises:
            ValueError: Raised if any header value is not string.
        """
        return await self._network_manager.set_extra_http_headers(headers)

    async def set_user_agent(
        self,
        user_agent: str,
        user_agent_metadata: str | None = None,
    ) -> None:
        """
        Update the user agent to be sent with every request.
        Wrapper for `Page.network_manager.set_user_agent`.

        Args:
            user_agent (str): User agent string.
            user_agent_metadata (str | None, optional): Experimental, used to
                specify user agent client hints to emulate.
        """
        return await self._network_manager.set_user_agent(
            user_agent,
            user_agent_metadata,
        )

    async def metrics(self) -> dict[str, Any]:
        """
        Return dictionary with current values of runtime metrics.

        Returns:
            dict[str, Any]: Runtime metrics as dictionary.
        """
        self._is_firefox(caller=self)
        response = await self._client.send(PERFORMANCE_GET_METRICS)
        return self._build_metrics(response.get("metrics", []))

    async def content(self) -> str:
        """
        Get encoded string representation of HTML in this `Page.main_frame`.
        Wrapper for `Page.main_frame.content`.

        Returns:
            str: HTML content.
        """
        frame = self._ensure_frame()
        return await frame.content()

    async def set_content(self, html: str) -> None:
        """
        Set the value of the HTML on this `Page.main_frame`.
        Wrapper for `Page.main_frame.set_content`.

        This does not change the value of the "document" request's response,
        it only changes what is shown in the window.

        Response values can be changed by intercepting requests or responses,
        see `Page.on` for more.

        Args:
            html (str): HTML content to set to.
        """
        frame = self._ensure_frame()
        await frame.set_content(html)

    async def goto(
        self,
        url: str,
        timeout: int | None = None,
        wait_until: list[LIFECYCLE_EVENTS] | LIFECYCLE_EVENTS = "load",
    ) -> Response | None:
        """
        Navigate to the given `url`.

        Args:
            url (str): URL to navigate to. Should be fully-qualified, starting
                with "https://" or "http://" (allowed if `ignore_https_errors`).
            timeout (int | None, optional): Time in milliseconds to wait for
                `wait_until` event. Defaults to None.
            wait_until (list[LIFECYCLE_EVENTS] | LIFECYCLE_EVENTS, optional):
                Event(s) to wait for, can be one or some combination of
                "load", "domcontentloaded", "networkidle", and
                "networkalmostidle". Defaults to "load".

        Raises:
            PageError: Raised if error encountered during navigation.
            MokrTimeoutError: Raised if timeout exceeded before all expected
                `wait_until` events are recieved.

        Returns:
            Response | None: Last response recieved after all redirects, if any.
        """
        main_frame = self._ensure_frame()
        referrer = self._network_manager.extra_http_headers.get('referer', '')
        all_responses = {}
        event_listeners = [
            add_event_listener(
                self._network_manager,
                NETWORK_MGR_REQUEST_FINISHED,
                lambda response: self._stash_response(all_responses, response),
            )
        ]
        _timeout = timeout if timeout else self._default_navigation_timeout
        watcher = NavigationWaiter(
            self._frame_manager,
            main_frame,
            _timeout,
            wait_until,
        )
        result = await self._navigate(url, referrer)
        if result is not None:
            raise PageError(result)
        result = await watcher.navigation_promise()
        watcher.cancel()
        remove_event_listeners(event_listeners)
        error = result[0].pop().exception()
        if error:
            raise error
        responses = all_responses.get(main_frame._navigation_url, [])
        return responses[-1] if responses else None

    async def go_back(
        self,
        timeout: int | None = None,
        wait_until: list[LIFECYCLE_EVENTS] | LIFECYCLE_EVENTS = "load",
    ) -> Response | None:
        """
        Navigate back in history.

        Args:
            timeout (int | None, optional): Time in milliseconds to wait for
                `wait_until` event. Defaults to None.
            wait_until (list[LIFECYCLE_EVENTS] | LIFECYCLE_EVENTS, optional):
                Event(s) to wait for, can be one or some combination of
                "load", "domcontentloaded", "networkidle", and
                "networkalmostidle". Defaults to "load".

        Returns:
            Response | None: History `mokr.network.Response`, if any.
        """
        return await self._navigate_history(-1, timeout, wait_until)

    async def go_forward(
        self,
        timeout: int | None = None,
        wait_until: list[LIFECYCLE_EVENTS] | LIFECYCLE_EVENTS = "load",
    ) -> Response | None:
        """
        Navigate next in history.

        Args:
            timeout (int | None, optional): Time in milliseconds to wait for
                `wait_until` event. Defaults to None.
            wait_until (list[LIFECYCLE_EVENTS] | LIFECYCLE_EVENTS, optional):
                Event(s) to wait for, can be one or some combination of
                "load", "domcontentloaded", "networkidle", and
                "networkalmostidle". Defaults to "load".

        Returns:
            Response | None: History `mokr.network.Response`, if any.
        """
        return await self._navigate_history(1, timeout, wait_until)

    async def refresh(
        self,
        timeout: int | None = None,
        wait_until: list[LIFECYCLE_EVENTS] | LIFECYCLE_EVENTS = "load",
    ) -> Response | None:
        """
        Refresh (reload) the current page.

        Args:
            timeout (int | None, optional): Time in milliseconds to wait for
                `wait_until` event. Defaults to None.
            wait_until (list[LIFECYCLE_EVENTS] | LIFECYCLE_EVENTS, optional):
                Event(s) to wait for, can be one or some combination of
                "load", "domcontentloaded", "networkidle", and
                "networkalmostidle". Defaults to "load".

        Raises:
            MokrTimeoutError: Raised if timeout exceeded before all expected
                `wait_until` events are recieved.

        Returns:
            Response | None: Last response recieved after all redirects, if any.
        """
        response = (
            await asyncio.gather(
                self.wait_for_navigation(timeout, wait_until),
                self._client.send(PAGE_RELOAD),
            )
        )[0]
        return response

    async def wait_for_navigation(
        self,
        timeout: int | None = None,
        wait_until: list[LIFECYCLE_EVENTS] | LIFECYCLE_EVENTS = "load",
    ) -> Response | None:
        """
        Wait for navigation to occur. Navigation is triggered by `Page.goto`,
        `Page.refresh`, redirects, and history (`Page.go_back` and
        `Page.go_forward`).

        Note that anchor navigation will not and history navigation may not
        yield any `mokr.network.Response` objects.

        Args:
            timeout (int | None, optional): Time in milliseconds to wait for
                `wait_until` event. Defaults to None.
            wait_until (list[LIFECYCLE_EVENTS] | LIFECYCLE_EVENTS, optional):
                Event(s) to wait for, can be one or some combination of
                "load", "domcontentloaded", "networkidle", and
                "networkalmostidle". Defaults to "load".

        Raises:
            MokrTimeoutError: Raised if timeout exceeded before all expected
                `wait_until` events are recieved.

        Returns:
            Response | None: Last response recieved after all redirects, if any.
        """
        main_frame = self._ensure_frame()
        _timeout = timeout if timeout else self._default_navigation_timeout
        watcher = NavigationWaiter(
            self._frame_manager,
            main_frame,
            _timeout,
            wait_until,
        )
        all_responses = {}
        listener = add_event_listener(
            self._network_manager,
            NETWORK_MGR_REQUEST_FINISHED,
            lambda response: self._stash_response(all_responses, response),
        )
        result = await watcher.navigation_promise()
        remove_event_listeners([listener])
        error = result[0].pop().exception()
        if error:
            raise error
        responses = all_responses.get(main_frame._navigation_url, [])
        return responses[-1] if responses else None

    async def wait_for_request(
        self,
        url: str | None = None,
        method: Callable | None = None,
        timeout: int | None = 30000,
    ) -> Request:
        """
        Wait for a `mokr.network.Request` whose `url` matches the `url` or
        that evaluates to True when passed to `method`.

        Args:
            url (str | None, optional): URL to wait. Defaults to None.
            method (Callable | None, optional): Callable to pass
                `mokr.network.Request`s emitted by `Page.network_manager` to.
                Defaults to None.
            timeout (int | None, optional): Time in milliseconds to wait.
                Defaults to 30000.

        Raises:
            ValueError: Raised if neither `url` nor `method` given.

        Returns:
            Request: First matching `mokr.network.Request`.
        """
        if url:
            callback = lambda request: url == request.url
        elif method:
            callback = lambda request: bool(method(request))
        else:
            raise ValueError("Must provide url or method.")
        waiter = EventWaiter(
            self._network_manager,
            NETWORK_MGR_REQUEST,
            callback,
            timeout,
            self._client._loop,
        )
        return await waiter.wait()

    async def wait_for_response(
        self,
        url: str | None = None,
        method: Callable | None = None,
        timeout: int | None = 30000,
    ) -> Response:
        """
        Wait for a `mokr.network.Response` whose `url` matches the `url` or
        that evaluates to True when passed to `method`.

        Args:
            url (str | None, optional): URL to wait. Defaults to None.
            method (Callable | None, optional): Callable to pass
                `mokr.network.Response`s emitted by `Page.network_manager` to.
                Defaults to None.
            timeout (int | None, optional): Time in milliseconds to wait.
                Defaults to 30000.

        Raises:
            ValueError: Raised if neither `url` nor `method` given.

        Returns:
            Request: First matching `mokr.network.Response`.
        """
        if url:
            callback = lambda response: url == response.url
        elif method:
            callback = lambda response: bool(method(response))
        else:
            raise ValueError("Must provide url or method.")
        waiter = EventWaiter(
            self._network_manager,
            NETWORK_MGR_RESPONSE,
            callback,
            timeout,
            self._client._loop,
        )
        return await waiter.wait()

    async def bring_to_front(self) -> None:
        """Bring this `Page` to front (activate tab)."""
        await self._client.send(PAGE_BRING_TO_FRONT)

    async def set_javascript_enabled(self, choice: bool) -> None:
        """
        Enable or disable JavaScript on this `Page`.

        Args:
            choice (bool): True to enable, False to disable.
        """
        self._is_firefox(caller=self)
        if self._javascript_enabled == choice:
            return
        self._javascript_enabled = choice
        await self._client.send(
            EMULATION_DISABLE_SCRIPT_EXECUTION,
            {'value': not choice}
        )

    async def set_bypass_csp(self, choice: bool) -> None:
        """
        Enable or disable the target page's Content Security Policy.

        Note this should be called before navigation to the domain as this
        happens at CSP initialization.

        Args:
            choice (bool): True to enable, False to disable.
        """
        self._is_firefox(caller=self)
        await self._client.send(PAGE_SET_BYPASS_CSP, {'enabled': choice})

    async def emulate_media(
        self,
        media_type: Literal["screen", "print"] | None = None,
    ) -> None:
        """
        Emulate CSS media type on target page.

        Args:
            media_type (Literal["screen", "print"] | None, optional):
                Media type to emulate; one of "screen", "print", or None.
                Defaults to None.

        Raises:
            ValueError: _description_
        """
        self._is_firefox(caller=self)
        if media_type not in ['screen', 'print', None, '']:
            raise ValueError(f'Unsupported media type: {media_type}')
        await self._client.send(
            EMULATION_SET_EMULATED_MEDIA,
            {'media': media_type or ''},
        )

    async def set_viewport(self, viewport: dict) -> None:
        """
        Run viewport adjustments based off given `viewport` parameters.
        Not all viewport options are considered, only: "isMobile", "width",
        "height", "deviceScaleFactor", "isLandscape", and "hasTouch".

        Args:
            viewport (dict[str, bool  |  int]): The parameters to adjust
                the viewport to.
        """
        needs_reload = await self._viewport_manager.emulate_viewport(viewport)
        self._viewport = viewport
        if needs_reload:
            await self.reload()

    async def evaluate(
        self,
        page_function: str,
        *args: Any,
        force_expr: bool = False,
    ) -> dict | None:
        """
        Execute a JavaScript function with given arguments.
        Runs this `Page`'s default `mokr.frame.Frame.evaluate`.

        Args:
            page_function (str): JavaScript function to run.
            force_expr (bool): If True, treat `page_function` as an expression.
                Otherwise, automatically determine if it is a function or
                an expression. Defaults to False.

        Raises:
            `mokr.exceptions.NetworkError` if an unhandled error occurs
            either evaluating the function or requesting the resulting object.

        Returns:
            dict | None: The decoded object (dict) via
                `mokr.execution.JavascriptHandle.json` or None if a known
                error occurs decoding it.
        """
        frame = self._ensure_frame()
        return await frame.evaluate(page_function, *args, force_expr=force_expr)

    async def evaluate_on_new_document(
        self,
        page_function: str,
        *args: str,
    ) -> None:
        """
        Attach a function that will be evaluated on every new document.
        This occurs when a page or frame is navigated or attached.

        Args:
            page_function (str): JavaScript function to attach.
        """
        source = self._make_javascript_function_string(page_function, *args)
        await self._client.send(PAGE_ADD_SCRIPT_TO_EVAL, {'source': source})

    async def set_request_cache(self, choice: bool = True) -> None:
        """
        Enable or disable request caching. Request caching caches requests in
        the browser, not `mokr.network.Request` objects.

        Does not check if request caching was enabled already by
        `Page.network_manager`.

        Args:
            enabled (bool, optional): True to enable, False to disable.
                Defaults to True.
        """
        await self._client.send(
            NETWORK_CACHE_DISABLE,
            {'cacheDisabled': not choice},
        )

    async def screenshot(
        self,
        file_type: Literal["png", "jpeg"] | None = None,
        file_path: str | None = None,
        jpeg_quality: int | None = None,
        full_page: bool = False,
        clip: dict[str, int] | None = None,
        omit_background: bool = False,
        encoding: Literal["binary", "base64"] = "binary",
        scale: int | float = 1,
    ) -> bytes:
        """
        Take a screenshot of the page viewport.

        Args:
            file_type (Literal["png", "jpeg"] | None, optional): File type
                to save the image as. Defaults to "png" if None.
            file_path (str | None, optional): Path to save image to. If given
                without `file_type`,  the type will be inferred.
                Defaults to None.
            jpeg_quality (int | None, optional): JPEG quality, 0-100 (only
                applicable if type is or is inferred to be JPEG.
                Defaults to 100 if None.
            full_page (bool, optional): Take a screenshot of the entire page if
                True, otherwise use the current viewport. Defaults to False.
            clip (dict[str, int] | None, optional): Applicable viewport
                settings to crop by. Defaults to None.
            omit_background (bool, optional): Make the background transparent
                if True.
                Not supported on Firefox.
                Defaults to False.
            encoding (Literal["binary", "base64"], optional): Encoding type
                to return the image data as. Defaults to "binary".
            scale (int | float, optional): Image scale, 0-1. Defaults to 1.

        Raises:
            ValueError: Raised if `file_type` isn't given and can't be inferred
                as a supported type from `file_path`, if given.

        Returns:
            bytes: The image content as bytes or base64 encoded bytes.
        """
        screenshot_type = "png"
        if file_type:
            screenshot_type = file_type
        elif file_path:
            mime_type, _ = mimetypes.guess_type(file_path)
            if mime_type == 'image/png':
                screenshot_type = 'png'
            elif mime_type == 'image/jpeg':
                screenshot_type = 'jpeg'
            else:
                screenshot_type = f"{mime_type} (mimetype)"
        if screenshot_type not in ['png', 'jpeg']:
            raise ValueError(f'Unsupported screenshot type: {screenshot_type}')
        return await self._screenshot_task(
            screenshot_type,
            file_path,
            jpeg_quality,
            full_page,
            clip,
            omit_background,
            encoding,
            scale,
        )

    async def pdf(
        self,
        file_path: str | None = None,
        scale: int | float = 1,
        display_header_footer: bool = False,
        header_template: str = None,
        footer_template: str = None,
        print_background: bool = False,
        landscape: bool = False,
        page_ranges: str = None,
        paper_format: str = "letter",
        width: str = None,
        height: str = None,
        margin: dict = None,
        prefer_css_page_size: bool = False,
    ) -> bytes:
        """
        Generate a PDF of the current page.

        `Page.pdf` generates a pdf of the page with "print" css media. To
        generate a pdf with screen media, first call
        `page.emulate_media("screen")`.

        The generated PDF will have modified colors for printing; this can
        be overriden by setting "-webkit-print-color-adjust" to "exact" in
        the page's stylesheet.

        Args:
            file_path (str | None, optional): File path to write output to, if
                any. Defaults to None.
            scale (int | float, optional): Scale the PDF, 0-1. Defaults to 1.
            display_header_footer (bool, optional): Display header and footer.
                Defaults to False.
            header_template (str, optional): HTML template for the print header.
                Should be valid HTML markup with following HTML classes used to
                inject printing values into them:
                    date: formatted print date
                    title: document title
                    url: document location
                    pageNumber: current page number
                    totalPages: total pages in the document
                For example, `<span class=title></span>` would generate span
                containing the title.
                Defaults to None.
            footer_template (str, optional): Same as `header_template` but for
                the page footer. Defaults to None.
            print_background (bool, optional): Print background graphics.
                Defaults to False.
            landscape (bool, optional): Orient the PDF as landscape if True,
                otherwise as portrait. Defaults to False.
            page_ranges (str, optional): Page ranges to print, can be individual
                such as "7" or numerous such as "2-4,7,11-17".
                Defaults to all if None.
            paper_format (str, optional): Paper format, overrides width/height.
                Formats are:
                    "letter": 8.5in x 11in
                    "legal": 8.5in x 14in
                    "tabloid": 11in x 17in
                    "ledger": 17in x 11in
                    "a0": 33.1in x 46.8in
                    "a1": 23.4in x 33.1in
                    "a2": 16.5in x 23.4in
                    "a3": 11.7in x 16.5in
                    "a4": 8.27in x 11.7in
                    "a5": 5.83in x 8.27in
                    "a6": 4.13in x 5.83in
                Defaults to "letter".
            width (str, optional): Page width. Accepts units, defaults to pixels
                if not provided. Units:
                    "px": pixel
                    "in": inch
                    "cm": centimeter
                    "mm": millimeter
                Defaults to None.
            height (str, optional): Page width. Accepts the same units as width.
                Defaults to None.
            margin (dict, optional): Margin values, can include any combination
                of "top", "bottom", "left", and "right".
                Accepts the same unit values as `width` and `height`.
                Defaults to None.
            prefer_css_page_size (bool, optional): Any CSS "@page" size.
                Overrides values from `paper_format`, `width`, and `height`.
                Defaults to False.

        Raises:
            ValueError: Raised if unknown paper format given.

        Returns:
            bytes: File content as bytes.
        """
        paper_formats = dict(
            letter={'width': 8.5, 'height': 11},
            legal={'width': 8.5, 'height': 14},
            tabloid={'width': 11, 'height': 17},
            ledger={'width': 17, 'height': 11},
            a0={'width': 33.1, 'height': 46.8},
            a1={'width': 23.4, 'height': 33.1},
            a2={'width': 16.5, 'height': 23.4},
            a3={'width': 11.7, 'height': 16.5},
            a4={'width': 8.27, 'height': 11.7},
            a5={'width': 5.83, 'height': 8.27},
        )
        if not header_template:
            header_template = ''
        if not footer_template:
            footer_template = ''
        if not page_ranges:
            page_ranges = ''
        paper_width = 8.5
        paper_height = 11.0
        if paper_format:
            fmt = paper_formats.get(paper_format.lower())
            if not fmt:
                raise ValueError(f"Unknown paper format: {paper_format}")
            paper_width = fmt['width']
            paper_height = fmt['height']
        else:
            paper_width = self._convert_print_param(width) or paper_width
            paper_height = self._convert_print_param(height) or paper_height
        if margin:
            margin_top = self._convert_print_param(margin.get('top')) or 0
            margin_left = self._convert_print_param(margin.get('left')) or 0
            margin_bottom = self._convert_print_param(margin.get('bottom')) or 0
            margin_right = self._convert_print_param(margin.get('right')) or 0
        else:
            margin_top = margin_left = margin_bottom = margin_right = 0
        result = await self._client.send(
            PAGE_PRINT_TO_PDF,
            dict(
                landscape=landscape,
                displayHeaderFooter=display_header_footer,
                headerTemplate=header_template,
                footerTemplate=footer_template,
                printBackground=print_background,
                scale=scale,
                paperWidth=paper_width,
                paperHeight=paper_height,
                marginTop=margin_top,
                marginBottom=margin_bottom,
                marginLeft=margin_left,
                marginRight=margin_right,
                pageRanges=page_ranges,
                preferCSSPageSize=prefer_css_page_size,
            ),
        )
        buffer = base64.b64decode(result.get('data', b''))
        if file_path:
            with open(file_path, 'wb') as f:
                f.write(buffer)
        return buffer

    async def title(self) -> str:
        """
        Get the title for this `Page.main_frame`.
        Wrapper for `Page.main_frame.title`.

        Returns:
            str: Page title.
        """
        frame = self._ensure_frame()
        return await frame.title()

    async def close(self, run_before_unload: bool = False) -> None:
        """
        Close the remote page.

        Args:
            run_before_unload (bool, optional): Whether to run handlers
                registered to "Window.beforeunload".
                Defaults to False.

        Raises:
            PageError: Raised if the remote connection has already been closed.
        """
        conn = self._client._connection
        if conn is None:
            raise PageError(
                'Protocol Error: Connection Closed. '
                'Most likely the page has been closed.'
            )
        if run_before_unload:
            await self._client.send(PAGE_CLOSE)
        else:
            await conn.send(TARGET_CLOSE, {'targetId': self._target._targetId})
            await self._target._is_closed_promise

    async def click(
        self,
        selector: str,
        button: Literal["left", "right", "middle"] = "left",
        click_count: int = 1,
        delay: int | float | None = 1000,
    ) -> None:
        """
        Click the first element that matches `selector`.
        Wrapper for `Page.main_frame.click`.

        This method is a shortcut for running `Page.query_selector` and then
        running `mokr.execution.ElementHandle.click` on the resultant
        ElementHandle. In either case, an element will be scrolled into view
        if needed, and the center of it clicked with the bound `Page.mouse`.

        Args:
            selector (str): Selector to query element by.
            button (Literal["left", "right", "middle"], optional): Mouse button
                to click with. Defaults to "left".
            click_count (int, optional): Number of clicks to run. Defaults to 1.
            delay (int | float | None, optional): Time in milliseconds to wait
                before each click. Defaults to 1000.

        Raises:
            PageError: Raised if no element is found with given `selector`.
        """
        frame = self._ensure_frame()
        await frame.click(selector, button, click_count, delay)

    async def hover(self, selector: str) -> None:
        """
        Mouse hover over the first element that matches `selector`.
        Wrapper for `Page.main_frame.hover`.

        Raises:
            PageError: Raised if no element is found with given `selector`.

        Args:
            selector (str): Selector to query element by.
        """
        frame = self._ensure_frame()
        await frame.hover(selector)

    async def focus(self, selector: str) -> None:
        """
        Focus on the first element that matches `selector`.
        Wrapper for `Page.main_frame.focus`.

        Raises:
            PageError: Raised if no element is found with given `selector`.

        Args:
            selector (str): Selector to query element by.
        """
        frame = self._ensure_frame()
        await frame.focus(selector)

    async def select(self, selector: str, values: list[str]) -> list[str]:
        """
        Select options on a "select" element.
        Wrapper for `Page.main_frame.select`.

        Args:
            selector (str): Selector to query element by.
            values (list[str]): List of string options to select by.

        Returns:
            list[str]: List of selected values.
        """
        frame = self._ensure_frame()
        return await frame.select(selector, values)

    async def type_text(
        self,
        selector: str,
        text: str,
        delay: int | float = 0,
    ) -> None:
        """
        Focus on the first element that matches `selector` and type characters
        into it.
        Wrapper for `Page.main_frame.type_text`.

        Note that modifier keys do not alter text case, meaning sending
        `mokr.input.Keyboard.press("shift")` and typing
        `Page.type_text("input", "mokr")` will not type "MOKR" into the it.

        Raises:
            PageError: Raised if no element is found with given `selector`.

        Args:
            selector (str): Selector to query element by.
            text (str): Text to type.
            delay (int | float, optional): Time in milliseconds to wait between
                each character typed. Defaults to 0.
        """
        frame = self._ensure_frame()
        return await frame.type_text(selector, text, delay)

    def wait_for_timeout(self, timeout: int | float) -> Awaitable[None]:
        """
        Wait for the given amount of time. Same as `asyncio.sleep`.
        Wrapper for `Page.main_frame.wait_for_timeout`.

        Args:
            timeout (int | float): Time in milliseconds to wait.

        Returns:
            Awaitable[None]: Task to be awaited.
        """
        frame = self._ensure_frame()
        return frame.wait_for(timeout)

    def wait_for_selector(
        self,
        selector: str,
        visible: bool = False,
        hidden: bool = False,
        timeout: int = 30000,
    ) -> Awaitable[ElementHandle]:
        """
        Wait for element that matches `selector` to appear in DOM.
        If element is in DOM already when called, return immediately.
        Wrapper for `Page.main_frame.wait_for_selector`.

        Args:
            selector (str): Selector to query element by.
            visible (bool, optional): Element must also not be hidden.
                Defaults to False.
            hidden (bool, optional): Element must also be hidden.
                Defaults to False.
            timeout (int, optional): Time in milliseconds to wait.
                Defaults to 30000.

        Raises:
            MokrTimeoutError: Raised if timeout exceeded before element found.

        Returns:
            Awaitable[ElementHandle]: Newly created
                `mokr.execution.ElementHandle` from found object.
        """
        frame = self._ensure_frame()
        return frame.wait_for_selector(selector, visible, hidden, timeout)

    def wait_for_xpath(
        self,
        xpath: str,
        visible: bool = False,
        hidden: bool = False,
        timeout: int = 30000,
    ) -> Awaitable[ElementHandle]:
        """
        Wait for element that matches `xpath` expression to appear in DOM.
        If element is in DOM already when called, return immediately.
        Wrapper for `Page.main_frame.wait_for_xpath`.

        Args:
            xpath (str): Expression to query element by.
            visible (bool, optional): Element must also not be hidden.
                Defaults to False.
            hidden (bool, optional): Element must also be hidden.
                Defaults to False.
            timeout (int, optional): Time in milliseconds to wait.
                Defaults to 30000.

        Raises:
            MokrTimeoutError: Raised if timeout exceeded before element found.

        Returns:
            Awaitable[ElementHandle]: Newly created
                `mokr.execution.ElementHandle` from found object.
        """
        frame = self._ensure_frame()
        return frame.wait_for_xpath(xpath, visible, hidden, timeout)

    def wait_for_function(
        self,
        page_function: str,
        polling: Literal["raf", "mutation"] | int | float = "raf",
        timeout: int = 30000,
    ) -> Awaitable[JavascriptHandle]:
        """
        Wait until the given `page_function` returns a truthy value.
        Wrapper for `Page.main_frame.wait_for_function`.

        Args:
            page_function (str): JavaScript function to run.
            polling (Literal["raf", "mutation"] | int | float, optional):
                Polling type; if set to "raf", executes continously in
                "requestAnimationFrame", else if set to "mutation" executes
                only on DOM mutations.
                Defaults to "raf".
            timeout (int, optional): Time in milliseconds to wait.
                Defaults to 30000.

        Returns:
            Awaitable[JavascriptHandle]: JavascriptHandle from JavaScript
                `page_function` successful result.
        """
        frame = self._ensure_frame()
        return frame.wait_for_function(page_function, polling, timeout)

    async def fetch(
        self,
        url: str | None = None,
        timeout: int = None,
        request: Request | None = None,
        body: str | None = None,
        browsing_topics: bool | None = None,
        cache: Literal[
            "default",
            "no-store",
            "reload",
            "no-cache",
            "force-cache",
            "only-if-cache",
        ] | None = None,
        credentials: Literal["omit", "same-origin", "include"] | None = None,
        headers: dict | None = None,
        method: str = "GET",
        mode: Literal["cors", "no-cors", "same-origin"] = None,
        priority: Literal["high", "low", "auto"] | None = None,
        redirect: Literal["follow", "error", "manual"] | None = None,
        referrer: str | None = None,
        referrer_policy: Literal[
            "no-referrer",
            "no-referrer-when-downgrade",
            "same-origin",
            "origin",
            "strict-origin",
            "origin-when-cross-origin",
            "strict-origin-when-cross-origin",
            "unsafe-url",
        ] | None = None,
    ) -> Response:
        """
        Send a fetch request.
        Wrapper for `Page.fetch_domain.fetch`.

        Note that fetch is sensitive to the current site's CORS settings.
        Additionally, trying to send a fetch request before the DOM loads can
        hang forever.

        Args:
            url (str | None, optional): The url to request.
                Ignored if `request` is given.
                Defaults to None.
            timeout (int, optional): Time in milliseconds to wait.
                Defaults to None.
            request (Request | None, optional): A request object to pull `url`,
                `method`, and `headers` from. Defaults to None.
            body (str | None, optional): Body to add to the request.
                Defaults to None.
            browsing_topics (bool | None, optional): If True, selected topics
                will be sent in a Sec-Browsing-Topics header with the associated
                request.
                Defaults to None.
            cache (Literal[str]): How the request should
                interact with the HTTP cache. One of: "default", "no-store",
                "reload", "no-cache", "force-cache", or "only-if-cache".
                Defaults to None.
            credentials (Literal[str] | None, optional): How to handle
                credentials. One of "omit", "same-origin", or "include".
                Defaults to None.
            headers (dict | None, optional): Headers to pass with the request.
                Defaults to None.
            method (str, optional): HTTP method. Defaults to "GET".
            mode (Literal["cors", "no-cors", "same-origin"], optional):
                CORS mode to use. Defaults to None.
            priority (Literal["high", "low", "auto"] | None, optional):
                Priority of request relative to others. Defaults to None.
            redirect (Literal["follow", "error", "manual"] | None, optional):
                How to handle redirects. Defaults to None ("follow").
            referrer (str | None, optional): Request referrer. Defaults to None.
            referrer_policy (Literal[str]): The referrer policy to use.
                One of "no-referrer", "no-referrer-when-downgrade",
                "same-origin", "origin", "strict-origin",
                "origin-when-cross-origin", "strict-origin-when-cross-origin",
                or "unsafe-url".
                Defaults to None ("strict-origin-when-cross-origin").

        Raises:
            ValueError: Raised if neither `url` nor `request` given.
            NetworkError: Raised if no response found when request resolves.

        Returns:
            Response: Last response recieved after all redirects, if any.
        """
        return await self._fetch_domain.fetch(
            url,
            timeout,
            request,
            body,
            browsing_topics,
            cache,
            credentials,
            headers,
            method,
            mode,
            priority,
            redirect,
            referrer,
            referrer_policy
        )

    async def send(
        self,
        url: str,
        method: HTTP_METHODS = "GET",
        params: dict | None = None,
        headers: dict | None = None,
        data: dict | None = None,
        json: dict | None = None,
        **httpx_request_kwargs,
    ) -> Response:
        """
        Send an HTTP request.
        Wrapper for `Page.http_domain.send`.

        Args:
            url (str): The url to request.
            method (Literal[str], optional): HTTP method. Defaults to "GET".
            params (dict | None, optional): Parameters to encode in the URL
                before sending. Defaults to None (omitted).
            headers (dict | None, optional): Additional headers to send with
                this request beyond what was set when the client was created.
                Defaults to None (omitted).
            data (dict | None, optional): Data payload to deliver with request
                (form-encoded data). Only for "PUT", "POST", and "PATCH".
                Defaults to None (omitted).
            json (dict | None, optional): Data payload to deliver with request
                (JSON-encoded data). Only for "PUT", "POST", and "PATCH".
                Defaults to None (omitted).

        Raises:
            ValueError: Raised if invalid HTTP method given or `data` or `json`
                arguments given with incompatible methods.

        Returns:
            Response: A `mokr.network.Response` created from the
                `httpx.Response`. All redirects will also be created, and
                original requests and responses will be accessible from the
                `mokr.network` objects.
        """
        return await self.http_domain.send(
            url,
            method,
            params,
            headers,
            data,
            json,
            **httpx_request_kwargs,
        )



from mokr.connection import DevtoolsConnection
from mokr.constants import INPUT_TOUCH
from mokr.input.keyboard import Keyboard


class Touchscreen():
    def __init__(
        self,
        client: DevtoolsConnection,
        keyboard: Keyboard,
    ) -> None:
        """
        Class to emulate touchscreen.

        Args:
            client (DevtoolsConnection): Remote `DevtoolsConnection` instance.
            keyboard (Keyboard): Active `mokr.input.Keyboard` from the parent
                `mokr.input.Keyboard`. Passes active modifiers such as CTRL.
        """
        self._client = client
        self._keyboard = keyboard

    async def tap(self, x: float, y: float) -> None:
        """
        Send touch down and up events to the center of the target coordinates.

        Args:
            x (float): X coordinate to move to.
            y (float): Y coordinate to move to.
        """
        touch_points = [{'x': round(x), 'y': round(y)}]
        await self._client.send(
            INPUT_TOUCH,
            {
                'type': 'touchStart',
                'touchPoints': touch_points,
                'modifiers': self._keyboard._modifiers,
            },
        )
        await self._client.send(
            INPUT_TOUCH,
            {
                'type': 'touchEnd',
                'touchPoints': [],
                'modifiers': self._keyboard._modifiers,
            },
        )



import re
import os
import platform
import shutil
import stat
import subprocess
import sys
import tarfile
import tempfile
from io import BytesIO
from pathlib import Path
from zipfile import ZipFile

import requests
from tqdm import tqdm

from mokr.constants import CHROME_VERSION, FIREFOX_BUILD, INSTALL_PATH


CHROME_DL_HOST = f'https://storage.googleapis.com/chrome-for-testing-public/{CHROME_VERSION}'  # noqa
FIREFOX_DL_HOST = f"https://archive.mozilla.org/pub/firefox/nightly/2024/01/2024-01-21-20-40-46-mozilla-central/firefox-{FIREFOX_BUILD}"  # noqa

CR_DOWNLOAD_URLS = {
    'linux': f'{CHROME_DL_HOST}/linux64/chrome-linux64.zip',
    'darwin': f'{CHROME_DL_HOST}/mac-x64/chrome-mac-x64.zip',
    'mac_arm': f'{CHROME_DL_HOST}/mac-arm64/chrome-mac-arm64.zip',
    'win32': f'{CHROME_DL_HOST}/win64/chrome-win64.zip',
}
FF_DOWNLOAD_URLS = {
    "linux": f"{FIREFOX_DL_HOST}.en-US.linux-x86_64.tar.bz2",
    "darwin": f"{FIREFOX_DL_HOST}.en-US.mac.dmg",
    "win32": f"{FIREFOX_DL_HOST}.en-US.win64.zip",
}

CR_BINARY_NAMES = {
    'linux': INSTALL_PATH / CHROME_VERSION / 'chrome-linux64' / 'chrome',
    'darwin': INSTALL_PATH / CHROME_VERSION / 'chrome-mac-x64' / 'Chrome.app' / 'Contents' / 'MacOS' / 'Chrome',  # noqa
    'mac_arm': INSTALL_PATH / CHROME_VERSION / 'chrome-mac-arm64' / 'Google chrome for Testing.app' / 'Contents' / 'MacOS' / 'Google Chrome for Testing',  # noqa
    'win32': INSTALL_PATH / CHROME_VERSION / 'chrome-win64' / 'chrome.exe',
}
FF_BINARY_NAMES = {
    "linux": INSTALL_PATH / FIREFOX_BUILD / 'firefox' / 'firefox',
    "darwin": INSTALL_PATH / FIREFOX_BUILD / 'Firefox Nightly.app' / 'Contents' / 'MacOS' / 'firefox',  # noqa
    "win32": INSTALL_PATH / FIREFOX_BUILD / 'firefox' / 'firefox.exe',
}
CR_BINARY_PATHS = {
    k: v.expanduser().absolute() for k, v in CR_BINARY_NAMES.items()
}
FF_BINARY_PATHS = {
    k: v.expanduser().absolute() for k, v in FF_BINARY_NAMES.items()
}


def get_platform() -> str:
    """Small utility to get different URL for MacOS Silicon machines."""
    machine = platform.machine()
    if sys.platform == "darwin" and machine == "arm64":
        return "mac_arm"
    else:
        return sys.platform


def download_zip(browser_type: str, url: str) -> BytesIO:
    """
    Download browser from the given `url`.

    Args:
        browser_type (str): Target browser type, of "chrome" or "firefox".
        url (str): URL to download from.

    Raises:
        requests.HTTPError: Raised if response is bad (status code over 399).

    Returns:
        BytesIO: The browser content as a BytesIO object.
    """
    print(f'Starting {browser_type.title()} download.')
    data = BytesIO()
    with requests.get(url, stream=True) as response:
        if response.status_code >= 400:
            raise requests.HTTPError(f"Bad response from server at: {url}")
        try:
            total_length = int(response.headers['content-length'])
        except (KeyError, ValueError, AttributeError):
            total_length = 0
        process_bar = tqdm(total=total_length, unit_scale=True, unit='b')
        for chunk in response.iter_content(chunk_size=8192):
            data.write(chunk)
            process_bar.update(len(chunk))
        process_bar.close()
    return data


def browser_binary(browser_type: str) -> Path:
    """
    Get path of the system target browser binary.

    Args:
        browser_type (str): Target browser type, of "chrome" or "firefox".

    Returns:
        Path: Browser binary Path object.
    """
    types_to_maps = {
        "chrome": CR_BINARY_PATHS,
        "firefox": FF_BINARY_PATHS,
    }
    mapping = types_to_maps[browser_type]
    binary = mapping.get(get_platform()) or mapping[sys.platform]
    return binary


def ensure_binary(browser_type: str) -> bool:
    """
    Ensure the target browser exists on this machine.

    Args:
        browser_type (str): Target browser type, of "chrome" or "firefox".

    Returns:
        bool: True if binary exists, otherwise False.
    """
    return browser_binary(browser_type).exists()


def install_dmg(process_bar: tqdm, dmg_path: str, dest_path: Path) -> None:
    """
    Install browser from dmg file. Mounts the dmg and copies to target path.

    Args:
        process_bar (tqdm): A process bar to update as actions occur.
        dmg_path (str): Path to dmg file.
        dest_path (Path): Path to final directory.

    Raises:
        OSError: Raised if mount fails.
        FileNotFoundError: Raised if dmg contains no .app files.
    """
    process_bar.update(2)
    proc = subprocess.run(
        ["hdiutil", "attach", "-nobrowse", "-noautoopen", dmg_path],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    process_bar.update(5)
    stdout = proc.stdout.decode()
    volumes = re.search(r"[\t]/Volumes/(.*)[\n]", stdout)
    if not volumes:
        raise OSError(f"Couldn't mount dmg: {stdout}")
    mount_path = Path(volumes[0].strip())
    process_bar.update(6)
    exception = None
    try:
        app_paths = list(mount_path.glob("**/*.app"))
        if not app_paths:
            raise FileNotFoundError(f"Failed to find an app in: {mount_path}")
        app_path = app_paths[0]
        dest_path = dest_path / app_path.name
        shutil.copytree(app_path, dest_path, dirs_exist_ok=True)
        process_bar.update(9)
    except Exception as e:
        exception = e
    finally:
        proc = subprocess.run(
            ['hdiutil', "detach", dmg_path, "-quiet"],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        if exception:
            raise type(exception)(str(exception))
    process_bar.update(10)


def extract(browser_type: str, data: BytesIO, path: Path, url: str) -> None:
    """
    Extract given loaded zip file to target `path`.

    Args:
        browser_type (str): Target browser type, of "chrome" or "firefox".
        data (BytesIO): Zip file as BytesIO.
        path (Path): Destination path to extract to.
        url (str): Download URL the file came from.

    Raises:
        IOError: Raised if an unhandled error occurs and the browser simply
            doesn't end up extracted to the target.
    """
    # On mac, Firefox is a dmg file, need to mount it and copy contents.
    print(f'Beginning {browser_type.title()} extraction.')
    if url.lower().endswith(".dmg"):
        process_bar = tqdm(range(11))
        temp_dmg = tempfile.NamedTemporaryFile()
        temp_dmg.write(data.getbuffer())
        process_bar.update(1)
        try:
            install_dmg(process_bar, temp_dmg.name, path)
        except Exception:
            temp_dmg.close()
            raise
        process_bar.close()
    elif url.lower().endswith(".tar.bz2"):
        temp_bz = tempfile.NamedTemporaryFile()
        temp_bz.write(data.getbuffer())
        try:
            with open(temp_bz.name, "rb") as fileobj:
                total_bytes = os.stat(temp_bz.name).st_size
                process_bar = tqdm(total=total_bytes, unit_scale=True, unit='b')
                with tarfile.open(fileobj=fileobj, mode="r:bz2") as tar:
                    last = 0
                    for member in tar:
                        member_data = tar.extractfile(member)
                        if member_data is None:
                            continue
                        subpath = (path / member.name)
                        subpath.parent.mkdir(parents=True, exist_ok=True)
                        subpath.write_bytes(member_data.read())
                        process_bar.update(fileobj.tell() - last)
                        last = fileobj.tell()
                process_bar.close()
        except Exception:
            temp_bz.close()
            raise
    else:
        with ZipFile(data) as zf:
            for member in tqdm(zf.infolist()):
                zf.extract(member, str(path))
    if not ensure_binary(browser_type):
        raise IOError('Failed to extract browser.')
    # On MacOS and Linux, helpers need execute too or process crashes instantly.
    if sys.platform == "win32":
        browser_binary(browser_type).chmod(
            browser_binary(browser_type).stat().st_mode | stat.S_IXOTH
            | stat.S_IXGRP | stat.S_IXUSR | stat.S_IEXEC
        )
    else:
        for subpath in path.glob("**/*"):
            subpath.chmod(
                subpath.stat().st_mode | stat.S_IXOTH
                | stat.S_IXGRP | stat.S_IXUSR | stat.S_IEXEC
            )
    print(f'{browser_type.title()} successfully extracted to: {path}')


def install_binary(browser_type: str) -> None:
    """
    Download and extract binary.

    Args:
        browser_type (str): Target browser type, of "chrome" or "firefox".
    """
    type_to_infos = {
        "chrome": (CR_DOWNLOAD_URLS, CHROME_VERSION),
        "firefox": (FF_DOWNLOAD_URLS, FIREFOX_BUILD),
    }
    download_urls, version = type_to_infos[browser_type]
    download_url = (
        download_urls.get(get_platform()) or download_urls[sys.platform]
    )
    extract(
        browser_type,
        download_zip(browser_type, download_url),
        INSTALL_PATH / version,
        download_url,
    )



import os
import sys
from pathlib import Path


MOKR_VERSION = "0.1.1"
DEFAULT_CHROME_VERSION = "122.0.6261.128"
DEFAULT_FIREFOX_BUILD = "123.0a1"
CHROME_VERSION = os.environ.get(
    'MOKR_CHROME_VERSION',
    DEFAULT_CHROME_VERSION,
)
FIREFOX_BUILD = os.environ.get(
    'MOKR_CHROME_VERSION',
    DEFAULT_FIREFOX_BUILD,
)

if os.environ.get("MOKR_INSTALL_DIR"):
    INSTALL_DIR_NAME = os.environ["MOKR_INSTALL_DIR"]
elif sys.platform == "linux":
    INSTALL_DIR_NAME = "~/.cache/"
elif sys.platform == "darwin":
    INSTALL_DIR_NAME = "~/Library/Caches/"
elif sys.platform == "win32":
    INSTALL_DIR_NAME = os.path.expandvars("%USERPROFILE%/AppData/Local/")
else:
    raise NotImplementedError(
        "Unsupported platform. Set MOKR_INSTALL_DIR environment variable to"
        " bypass this check."
    )
INSTALL_PATH = Path(INSTALL_DIR_NAME).expanduser() / 'mokr'



from mokr.execution import JavascriptHandle


class ConsoleMessage():
    def __init__(
        self,
        kind: str,
        text: str,
        args: list[JavascriptHandle] = None
    ) -> None:
        """
        Representation of console messages, dispatched on `console` event in
        `mokr.browser.Page`.

        Args:
            kind (str): The type of message.
            text (str): The message body.
            args (list[JavascriptHandle], optional): Arguments attached to this
                message. Defaults to None.
        """
        self._type = kind
        self._text = text
        self._args = args if args is not None else []

    @property
    def kind(self) -> str:
        """Return type of this message."""
        return self._type

    @property
    def text(self) -> str:
        """Return text representation of the message body."""
        return self._text

    @property
    def args(self) -> list[JavascriptHandle]:
        """Return list of argumentss (JavascriptHandle) of this message."""
        return self._args



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Callable

from mokr.connection import DevtoolsConnection
from mokr.constants import (
    NETWORK_MGR_REQUEST,
    NETWORK_MGR_REQUEST_FINISHED,
    NETWORK_MGR_RESPONSE,
    NETWORK_REQUEST_WILL_BE_SENT,
    NETWORK_RESPONSE_RECVD,
)
from mokr.frame import FrameManager
from mokr.network.manager.base import NetworkManager
from mokr.network.request import Request
from mokr.network.response import Response

if TYPE_CHECKING:
    from mokr.browser.page import Page


LOGGER = logging.getLogger(__name__)


class FirefoxNetworkManager(NetworkManager):
    def __init__(
        self,
        page: Page,
        client: DevtoolsConnection,
        frame_manager: FrameManager,
        interception_callback_chain: list[Callable],
    ) -> None:
        """
        Class to handle requests in a given `mokr.browser.Page`.

        Note that functionality of this class is severely limited compared to
        it's Chrome counterpart, `mokr.network.ChromeNetworkManager`.

        Request interception does not block requests and can only be used to
        observe partial request and response data.

        The `interception_callback_chain` this class receives from the parent
        `mokr.browser.Page` is shared with spawned `mokr.network.Request`
        objects. While the request interception callback chain is run from the
        `Request` object, it is initiated during select events from this class.

        Args:
            page (Page): Parent `mokr.browser.Page`.
            client (DevtoolsConnection): A `mokr.connection.DevtoolsConnection`
                spawned by the parent `mokr.browser.Page`.
            frame_manager (FrameManager): The `mokr.frame.FrameManager`
                from the `mokr.browser.Page` that spawned this element.
            interception_callback_chain (list[Callable]): A list of callbacks
                to be passed into new `mokr.network.Request`s that will be run
                during "request" event interception, by that object. Inherited
                from the parent `mokr.browser.Page`.
        """

        self._events_to_methods = {
            NETWORK_REQUEST_WILL_BE_SENT: (
                lambda event: self._client._loop.create_task(
                    self._on_request(event)
                )
            ),
            NETWORK_RESPONSE_RECVD: self._on_response_received,
        }
        super().__init__(
            page,
            client,
            frame_manager,
            interception_callback_chain,
        )
        self._request_id_to_request = {}

    async def _on_request(self, event: dict) -> None:
        # Firefox request event never has a "redirectResponse", not handling.
        mokr_request_uuid = self._inspect_stack_for_mokr_uuid(event)
        requestId = event['requestId']
        frame = (
            self._frame_manager.frame(event['frameId'])
            if event.get('frameId') else None
        )
        is_navigation_request = bool(
            event.get('requestId') == event.get('loaderId')
            and event.get('type') == 'Document'
        )
        request = Request(
            self._page,
            self._client,
            event['requestId'],
            None,
            is_navigation_request,
            self._user_request_interception_enabled,
            event.get('request', {}).get('url'),
            event.get('type', ''),
            event.get('request', {}),
            frame,
            [],
            self._interception_callback_chain,
        )
        if mokr_request_uuid:
            request._mokr_request_uuid = mokr_request_uuid
        self._request_id_to_request[requestId] = request
        self.emit(NETWORK_MGR_REQUEST, request)
        await request._finalize_interceptions()

    def _on_response_received(self, event: dict) -> None:
        # Firefox doesn't emit on extraInfo event ever so not handling.
        request = self._request_id_to_request.get(event['requestId'])
        # FileUpload sends a response without a matching request.
        if not request:
            return
        _response = event.get('response', {})
        response = Response(
            self._client,
            request,
            _response.get('status', 0),
            _response.get('headers', {}),
            _response.get('fromDiskCache'),
            _response.get('fromServiceWorker'),
            True,
            _response.get('securityDetails'),
            None,
        )
        request._response = response
        self.emit(NETWORK_MGR_RESPONSE, response)
        # Hacky but Firefox doesn't emit a loading finished event.
        # Assume finished Document request means loading finished.
        if event["type"] == "Document":
            self.emit(NETWORK_MGR_REQUEST_FINISHED, request)

    @property
    def extra_http_headers(self) -> dict[str, str]:
        """
        For parity; Firefox doesn't support sending extra headers via CDP.
        """
        return self._extra_http_headers

    async def set_request_interception(self, *args, **kwargs) -> None:
        """
        Not supported by Firefox.

        Raises:
            FirefoxNotImplementedError: When Firefox unsupported errors are on.
        """
        self._page._is_firefox(caller=self)

    async def set_extra_http_headers(self, *args, **kwargs) -> None:
        """
        Not supported by Firefox.

        Raises:
            FirefoxNotImplementedError: When Firefox unsupported errors are on.
        """
        self._page._is_firefox(caller=self)
